package com.parse;

import android.os.AsyncTask;

/**
 * The BackgroundTask makes it easier to do something in the background.
 * <p>
 * Clients should not have to create a BackgroundTask; they are just
 * created by our asynchronous code.
 * <p>
 * To use a BackgroundTask, create it with the ParseCallback that you
 * want to run in the UI thread after the task is complete. Then override the
 * run() function with the code you want to run in the background thread. Then
 * execute the task.
 */
abstract class BackgroundTask<T> extends AsyncTask<Void, Void, Void> {
	// Can be null if no callback is needed.
	private ParseCallback<T> callback;
	
	// The result of the operation.
	private T result;
	
	// Any exception that occurred during the operation.
	private ParseException exception;
	
	BackgroundTask(ParseCallback<T> theCallback) {
		super();
		result = null;
		exception = null;
		callback = theCallback;
	}

	// The library code should override this.
	abstract public T run() throws ParseException;
	
	@Override
	protected Void doInBackground(Void... params) {
		try {
			result = run();
			return null;
		} catch (ParseException e) {
			exception = e;
			return null;
		}
	}
	
	@Override
	protected void onPostExecute(Void v) {
		if (callback != null) {
			callback.internalDone(result, exception);
		}
	}
	

	
	// Just executes the task in this thread.
	// Either use this, or execute(), not both.
	void executeInThisThread() {
		doInBackground();
		onPostExecute(null);
	}
	
	// A helper to execute a task.
	// This only exists so it can be mocked statically.
	static int executeTask(BackgroundTask<?> task) {
		task.execute();
		return 0;
	}
}
