package com.parse;

/**
 * A CountCallback is used to run code after a {@link ParseQuery} is used to count
 * objects matching a query in a background thread.
 * <p>
 * The easiest way to use a GetCallback is through an anonymous inner class. Override the
 * <code>done</code> function to specify what the callback should do after the count is
 * complete. The <code>done</code> function will be run in the UI thread, while the count
 * happens in a background thread.
 * This ensures that the UI does not freeze while the fetch happens. 
 * <p>
 * For example, this sample code counts objects of class <code>"MyClass"</code>.
 * It calls a different function depending on whether the count succeeded or not.
 * <p>
 * <pre>
 * ParseQuery query = new ParseQuery("MyClass");
 * query.countInBackground(new CountCallback() {
 *     public void done(int count, ParseException e) {
 *         if (e == null) {
 *             objectsWereCountedSuccessfully(count);
 *         } else {
 *             objectCountingFailed();
 *         }
 *     }
 * });
 * </pre>
 */
public abstract class CountCallback extends ParseCallback<Integer> {
	/**
	 * Override this function with the code you want to run after the
	 * count is complete.
	 * @param count The number of objects matching the query, or -1 if it failed.
	 * @param e The exception raised by the count, or null if it succeeded.
	 */
	public abstract void done(int count, ParseException e);
	 
	@Override
	void internalDone(Integer returnValue, ParseException e) {
		if (e == null) {
			done(returnValue.intValue(), null);
		} else {
			done(-1, e);			
		}
	}

}
