package com.parse;

/**
 * A DeleteCallback is used to run code after saving a {@link ParseObject} in a background thread.
 * <p>
 * The easiest way to use a DeleteCallback is through an anonymous inner class. Override the
 * <code>done</code> function to specify what the callback should do after the delete is complete. The
 * <code>done</code> function will be run in the UI thread, while the delete happens in a background thread.
 * This ensures that the UI does not freeze while the delete happens. 
 * <p>
 * For example, this sample code deletes the object <code>myObject</code> and calls a different function
 * depending on whether the save succeeded or not.
 * <p>
 * <pre>
 * myObject.deleteInBackground(new DeleteCallback() {
 *     public void done(ParseException e) {
 *         if (e == null) {
 *             myObjectWasDeletedSuccessfully();
 *         } else {
 *             myObjectDeleteDidNotSucceed();
 *         }
 *     }
 * });
 * </pre>
 */
public abstract class DeleteCallback extends ParseCallback<Void> {
	
	/**
	 * Override this function with the code you want to run after
	 * the delete is complete.
	 * @param e The exception raised by the delete, or null if it succeeded.
	 */
	public abstract void done(ParseException e);
	
	@Override
	final void internalDone(Void returnValue, ParseException e) {
		done(e);
	}

}
