package com.parse;

import java.util.List;

/**
 * A FindCallback is used to run code after a {@link ParseQuery} is used to fetch a list
 * of {@link ParseObject}s in a background thread.
 * <p>
 * The easiest way to use a FindCallback is through an anonymous inner class. Override the
 * <code>done</code> function to specify what the callback should do after the fetch is complete. The
 * <code>done</code> function will be run in the UI thread, while the fetch happens in a background thread.
 * This ensures that the UI does not freeze while the fetch happens. 
 * <p>
 * For example, this sample code fetches all objects of class <code>"MyClass"</code>.
 * It calls a different function depending on whether the fetch succeeded or not.
 * <p>
 * <pre>
 * ParseQuery query = new ParseQuery("MyClass");
 * query.findInBackground(new FindCallback() {
 *     public void done(List&lt;ParseObject&gt; objects, ParseException e) {
 *         if (e == null) {
 *             objectsWereRetrievedSuccessfully(objects);
 *         } else {
 *             objectRetrievalFailed();
 *         }
 *     }
 * });
 * </pre>
 */
public abstract class FindCallback extends ParseCallback<List<ParseObject>> {
	
	/**
	 * Override this function with the code you want to run after
	 * the fetch is complete.
	 * @param objects The objects that were retrieved, or null if it did not succeed.
	 * @param e The exception raised by the save, or null if it succeeded.
	 */
	public abstract void done(List<ParseObject> objects, ParseException e);
	
	@Override
	final void internalDone(List<ParseObject> returnValue, ParseException e) {
		done(returnValue, e);
	}
}
