package com.parse;

/**
 * A LogInCallback is used to run code after logging in a user.
 * <p>
 * The easiest way to use a LogInCallback is through an anonymous inner class. Override the
 * <code>done</code> function to specify what the callback should do after the login is complete. The
 * <code>done</code> function will be run in the UI thread, while the login happens in a background thread.
 * This ensures that the UI does not freeze while the save happens. 
 * <p>
 * For example, this sample code logs in a user and calls a different function
 * depending on whether the login succeeded or not.
 * <p>
 * <pre>
 * ParseUser.logInInBackground("username", "password", new LogInCallback() {
 *     public void done(ParseUser user, ParseException e) {
 *         if (e == null && user != null) {
 *             loginSuccessful();
 *         } else if (user == null) {
 *         	   usernameOrPasswordIsInvalid();
 *         } else {
 *             somethingWentWrong();
 *         }
 *     }
 * });
 * </pre>
 */
public abstract class LogInCallback extends ParseCallback<ParseUser> {
	
	/**
	 * Override this function with the code you want to run after 
	 * the save is complete.
	 * @param user The user that logged in, if the username and password is valid.
	 * @param e The exception raised by the login.
	 */
	public abstract void done(ParseUser user, ParseException e);

	@Override
	void internalDone(ParseUser returnValue, ParseException e) {
		done(returnValue, e);
	}

}
