package com.parse;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.RandomAccessFile;
import java.io.UnsupportedEncodingException;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.Arrays;
import java.util.Collection;
import java.util.Comparator;
import java.util.Date;
import java.util.Iterator;
import java.util.SimpleTimeZone;

import org.json.JSONArray;
import org.json.JSONException;
import org.json.JSONObject;
import org.json.JSONTokener;

import android.content.Context;
import android.util.Log;

import com.parse.codec.binary.Base64;

/**
 * The Parse class contains static functions that handle global configuration for the
 * Parse library.
 */
public class Parse {
	private static final String TAG = "com.parse.Parse";
	static Context applicationContext;
	static String applicationId;
	static String clientKey;
	
	// We limit the size of ParseFile data to be 10mb.
	static int maxParseFileSize = 10 * 1048576;
	
	// We limit the cache to 2MB because that's about what the default browser uses.
	static int maxKeyValueCacheBytes = 2 * 1024 * 1024;
	
	// We limit to 1000 cache files to avoid taking too long while scanning the cache
	static int maxKeyValueCacheFiles = 1000;
	
	// Suppress constructor to prevent subclassing
	private Parse() {
		throw new AssertionError();
	}
	
	// This is the currently used date format.
	// It is precise to the millisecond.
	private static final DateFormat dateFormat;
	static {
		DateFormat format = new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss.SSS'Z'");
		format.setTimeZone(new SimpleTimeZone(0, "GMT"));
		dateFormat = format;
	}
	
	/**
	 * Authenticates this client as belonging to your application.
	 * This must be called before your application can use the Parse library.
	 * The recommended way is to put a call to <code>Parse.initialize</code> in each
	 * of your <code>onCreate</code> methods. An example:
	 * 
	 * <pre>
	 * import android.app.Application;
     * import com.parse.Parse;
     *
	 * public class MyApplication extends Application {
	 *
     *     public void onCreate() {
     *         Parse.initialize(this, "your application id", "your client key");
     *     }
     *
     * }
	 * </pre>
	 * 
	 * @param context The active Context for your application.
	 * @param applicationId The application id provided in the Parse dashboard. 
	 * @param clientKey The client key provided in the Parse dashboard.
	 */
	public static void initialize(
			Context context, String applicationId, String clientKey) {
		// We don't use the context yet, but we will once we can trust the
		// ParseObject.initialize calls have been deprecated. This will let us cache
		// on disk for greater efficiency.
		Parse.applicationId = applicationId;
		Parse.clientKey = clientKey;
		if (context != null) {
			applicationContext = context.getApplicationContext();
		}
	}
	
	// Helper function for code that runs when we haven't called initialize
	static void initializeContextIfNeeded(Context context) {
		if (applicationContext == null) {
			applicationContext = context;
		}
	}
	
	// Creates a directory to keep Parse files in.
	synchronized static File getParseDir() {
		checkContext();
		return applicationContext.getDir("Parse", 0);
	}

	// Creates a directory to keep cache-type files in.
	// The operating system will automatically clear out these files first
	// when space gets low.
	synchronized static File getKeyValueCacheDir() {
		checkContext();
		File appCacheDir = applicationContext.getCacheDir();
		File parseCacheDir = new File(appCacheDir, "ParseKeyValueCache");
		if (parseCacheDir.isDirectory() || parseCacheDir.mkdir()) {
			return parseCacheDir;
		}
		throw new RuntimeException("could not create Parse cache directory");
	}
	
	// Saves a key-value pair to the cache
	static void saveToKeyValueCache(String key, String value) {
		File dir = getKeyValueCacheDir();
		File f = new File(getKeyValueCacheDir(), key);
		
		try {
			FileOutputStream out = new FileOutputStream(f);
			out.write(value.getBytes("UTF-8"));
			out.close();
		} catch (UnsupportedEncodingException e) {
		} catch (IOException e) {
		}
		
		// Check if we should kick out old cache entries
		File[] files = dir.listFiles();
		int numFiles = files.length;
		int numBytes = 0;
		for (File file : files) {
			numBytes += file.length();
		}
		if (numFiles > maxKeyValueCacheFiles || numBytes > maxKeyValueCacheBytes) {
			// We need to kick out some cache entries.
			// Sort oldest-first. We touch on read so mtime is really LRU.
			Arrays.sort(files, new Comparator<File>() {
				public int compare(File f1, File f2) {
					return Long.valueOf(f1.lastModified()).compareTo(f2.lastModified());
				}
			});
		
			for (File file : files) {
				numFiles--;
				numBytes -= file.length();
				file.delete();
				
				if (numFiles <= maxKeyValueCacheFiles && numBytes <= maxKeyValueCacheBytes) {
					break;
				}
			}
		}
	}
	
	// Clears a key from the cache if it's there. If it's not there, this is a no-op.
	static void clearFromKeyValueCache(String key) {
		File file = new File(getKeyValueCacheDir(), key);
		file.delete();
	}
	
	// Loads a value from the key-value cache.
	// Returns null if nothing is there.
	static String loadFromKeyValueCache(String key) {
		File file = new File(getKeyValueCacheDir(), key);
		if (!file.exists()) {
			return null;
		}
		
		// Update mtime to make the LRU work
		Date now = new Date();
		file.setLastModified(now.getTime());
		
		try {
			RandomAccessFile f = new RandomAccessFile(file, "r");
			byte[] bytes = new byte[(int) f.length()];
			f.readFully(bytes);
			f.close();
			return new String(bytes, "UTF-8");
		} catch (IOException e) {
			Log.d(TAG, "error reading from cache: " + e.getMessage());
			return null;
		}
	}
	
	// Returns null if the value does not exist or is not json
	static Object jsonFromKeyValueCache(String key) {
		String raw = Parse.loadFromKeyValueCache(key);
		if (raw == null) {
			return null;
		}
		JSONTokener tokener = new JSONTokener(raw);
		
		try {
			Object o = tokener.nextValue();
			return o;
		} catch (JSONException e) {
			Log.d(TAG, "corrupted cache for " + key);
			clearFromKeyValueCache(key);
			return null;
		}
	}
	
	static void checkInit() {
		if (applicationId == null) {
			throw new RuntimeException(
					"applicationId is null. " +
					"You must call Parse.initialize(context, applicationId, clientKey)"
					+ " before using the Parse library.");
		}		
		if (clientKey == null) {
			throw new RuntimeException(
					"clientKey is null. " +
					"You must call Parse.initialize(context, applicationId, clientKey)"
					+ " before using the Parse library.");
		}		
	}
	
	static void checkContext() {
		if (applicationContext == null) {
			throw new RuntimeException(
					"applicationContext is null. " +
					"You must call Parse.initialize(context, applicationId, clientKey)"
					+ " before using the Parse library.");
		}
	}
	
	static Object decodeJSONObject(Object object) {
		JSONObject jsonObject;
		if (object instanceof JSONObject) {
			jsonObject = (JSONObject) object;
		} else {
			return null;
		}
		
		String typeString = jsonObject.optString("__type");
		if (typeString == null) {
			return null;
		}
		
		if (typeString.equals("Date")) {
			String iso = jsonObject.optString("iso");
			return parseDate(iso);
		}
		
		if (typeString.equals("Bytes")) {
			String base64 = jsonObject.optString("base64");			
			return Base64.decodeBase64(base64);
		}
		
		if (typeString.equals("Pointer")) {
			return new ParsePointer(
					jsonObject.optString("className"),
					jsonObject.optString("objectId"));
		}
		
		if (typeString.equals("File")) {
			return new ParseFile(
					jsonObject.optString("name"),
					jsonObject.optString("url"));
		}
		
		if (typeString.equals("GeoPoint")) {
			double latitude, longitude;			
			try {
				latitude = jsonObject.getDouble("latitude");
				longitude = jsonObject.getDouble("longitude");
			} catch (JSONException e) {
				throw new RuntimeException(e.getMessage());
			}
			return new ParseGeoPoint(latitude, longitude);
		}
		
		if (typeString.equals("Object")) {
			ParseObject output = ParseObject.build(jsonObject.optString("className"));
			JSONObject nested = new JSONObject();
			try {
				nested.put("data", jsonObject);
			} catch (JSONException e) {
				throw new RuntimeException(e.getMessage());
			}
			output.mergeObject(nested, false);
			return output;
		}
		
		return null;
	}
	
	static JSONObject encodeJSONObject(Object object) {
		try {
			if (object instanceof Date) {
				return dateToObject((Date) object);
			}

			if (object instanceof byte[]) {
				JSONObject json = new JSONObject();
				json.put("__type", "Bytes");
				json.put("base64", Base64.encodeBase64String((byte[]) object));
				return json;
			}
			
			if (object instanceof ParsePointer) {
				return pointerToJSONObject((ParsePointer) object);
			}
			
			if (object instanceof ParseFile) {
				ParseFile file = (ParseFile) object;
				JSONObject json = new JSONObject();
				json.put("__type", "File");
				json.put("name", file.getName());
				return json;
			}
			
			if (object instanceof ParseGeoPoint) {
				ParseGeoPoint point = (ParseGeoPoint) object;
				JSONObject json = new JSONObject();
				json.put("__type", "GeoPoint");
				json.put("latitude", point.getLatitude());
				json.put("longitude", point.getLongitude());
				return json;
			}
			
			if (object instanceof ParseACL) {
				ParseACL acl = (ParseACL) object;
				return acl.toJSONObject();
			}
			
			if (object instanceof ParseIncrementOperation) {
				JSONObject json = new JSONObject();
				json.put("__op", "Increment");
				json.put("amount", ((ParseIncrementOperation) object).getAmount());
				return json;
			}
			
			if (object instanceof ParseDeleteOperation) {
				JSONObject json = new JSONObject();
				json.put("__op", "Delete");
				return json;
			}
			
		} catch (JSONException e) {
			throw new RuntimeException(e.getMessage());
		}
		
		return null;
	}
		
	static Object maybeEncodeJSONObject(Object value) {
		// If this object has a special encoding, encode it and return the
		// encoded object.  Otherwise, just return the original object.
		JSONObject json = encodeJSONObject(value);
		if (null == json) {
			return value; 
		}
		return json;
	}
	
	static Object maybeReferenceAndEncode(Object value) {
		// If this is a ParseObject, we map it to a ParsePointer.
		// We then handle special encodings if appropriate. 
		if (value instanceof ParseObject) {
			value = ((ParseObject) value).address();
			if (null == value) {
				// object that hasn't been saved.
				throw new RuntimeException("bad reference encountered");
			}
		}
		return maybeEncodeJSONObject(value);
	}
	
	static JSONObject pointerToJSONObject(ParsePointer ptr) {
		JSONObject json = new JSONObject();
		try {
			json.put("__type", "Pointer");
			json.put("className", ptr.className);
			json.put("objectId", ptr.objectId);
		} catch (JSONException e) {
			// This should not happen
			throw new RuntimeException(e.getMessage());
		}
		return json;
	}
		
	static synchronized Date parseDate(String dateString) {
		try {
			return dateFormat.parse(dateString);
		} catch (java.text.ParseException e) {
			// Should never happen
			Log.d(TAG, "could not parse date: " + dateString);
			return null;
		}
	}
	
	static synchronized String encodeDate(Date date) {
		return dateFormat.format(date);
	}
	
	static JSONObject dateToObject(Date date) {
		JSONObject object = new JSONObject();
		String iso = encodeDate(date);
		try {
			object.put("__type", "Date");
			object.put("iso", iso);
		} catch (JSONException e) {
			// This should not happen
			throw new RuntimeException(e.getMessage());
		}
		return object;
	}
	
	// A helper for nonugly iterating over JSONObject keys.
	static Iterable<String> keys(JSONObject object) {
		final JSONObject finalObject = object;
		return new Iterable<String>() {
			public Iterator<String> iterator() {
				@SuppressWarnings("unchecked")
				Iterator<String> answer = (Iterator<String>) finalObject.keys(); 
				return answer;
			}
		};
	}
	
	static boolean isContainerObject(Object object) {
		return (object instanceof JSONObject 
				|| object instanceof JSONArray
				|| object instanceof ParseACL
				|| object instanceof ParseGeoPoint);
	}
	
	static Number addNumbers(Number first, Number second) {
		if (first instanceof Double) {
			return first.doubleValue() + second.doubleValue();
		} else if (first instanceof Long) {
			return first.longValue() + second.longValue();
		} else if (first instanceof Float) {
			return first.floatValue() + second.floatValue();
		} else if (first instanceof Short) {
			return first.shortValue() + second.shortValue();
		} else if (first instanceof Byte) {
			return first.byteValue() + second.byteValue();
		} else {
			return first.intValue() + second.intValue();
		}
	}
	
	static String join(Collection<String> items, String delimiter) {
        StringBuffer buffer = new StringBuffer();
        Iterator<String> iter = items.iterator();
        if (iter.hasNext()) {
            buffer.append(iter.next());
            while (iter.hasNext()) {
                buffer.append(delimiter);
                buffer.append(iter.next());
            }
        }
        return buffer.toString();
	}
}
