package com.parse;

import org.json.JSONException;
import org.json.JSONObject;

/**
 * A ParseACL is used to control which users can access or modify a particular object.
 * Each {@link ParseObject} can have its own ParseACL. You can grant read and
 * write permissions separately to any number of users, and you can also grant permissions
 * to "the public" so that, for example, any user could read a particular object but
 * only a particular set of users could write to that object.
 */
public class ParseACL {
	private JSONObject permissionsById;
	
	/**
	 * Creates an ACL with no permissions granted.
	 */
	public ParseACL() {
		permissionsById = new JSONObject();
	}
	
	// Internally we expose the json object this wraps
	JSONObject toJSONObject() {
		return permissionsById;
	}
	
	// A helper for creating a ParseACL from the wire.
	// We iterate over it rather than just copying to permissionsById so that we can
	// ensure it's the right format.
	static ParseACL createACLFromJSONObject(JSONObject object) {
		ParseACL acl = new ParseACL();
		for (String userId : Parse.keys(object)) {
			JSONObject permissions;
			try {
				permissions = object.getJSONObject(userId);
			} catch (JSONException e) {
				throw new RuntimeException("could not decode ACL: " + e.getMessage());
			}
			for (String accessType : Parse.keys(permissions)) {
				acl.setAccess(accessType, userId, true);
			}
		}
		return acl;
	}
	
	/**
	 * Creates an ACL where only the provided user has access.
	 * @param owner The only user that can read or write objects governed by this ACL.
	 */
	public ParseACL(ParseUser owner) {
		this();
		setReadAccess(owner, true);
		setWriteAccess(owner, true);
	}
	
	// Helper for setting stuff
	private void setAccess(String accessType, String userId, boolean allowed) {
		try {
			JSONObject permissions = permissionsById.optJSONObject(userId);
			if (permissions == null) {
				if (!allowed) {
					// The user already doesn't have this permission, so no action is needed.
					return;
				}
				permissions = new JSONObject();
				permissionsById.put(userId, permissions);
			}
			
			if (allowed) {
				permissions.put(accessType, true);
			} else {
				permissions.remove(accessType);
				if (permissions.length() == 0) {
					permissionsById.remove(userId);
				}
			}
		} catch (JSONException e) {
			throw new RuntimeException("JSON failure with ACL: " + e.getMessage());
		}
	}
	
	/**
	 * Set whether the public is allowed to read this object.
	 */
	public void setPublicReadAccess(boolean allowed) {
		setReadAccess("*", allowed);
	}
	
	/**
	 * Set whether the public is allowed to write this object.
	 */
	public void setPublicWriteAccess(boolean allowed) {
		setWriteAccess("*", allowed);
	}
	
	/**
	 * Set whether the given user id is allowed to read this object.
	 */
	public void setReadAccess(String userId, boolean allowed) {
		if (userId == null) {
			throw new IllegalArgumentException("cannot setReadAccess for null userId");
		}
		setAccess("read", userId, allowed);
	}
	
	/**
	 * Set whether the given user id is allowed to write this object.
	 */
	public void setWriteAccess(String userId, boolean allowed) {
		if (userId == null) {
			throw new IllegalArgumentException("cannot setWriteAccess for null userId");
		}
		setAccess("write", userId, allowed);
	}
	
	/**
	 * Set whether the given user is allowed to read this object.
	 */
	public void setReadAccess(ParseUser user, boolean allowed) {
		if (user.getObjectId() == null) {
			throw new IllegalArgumentException(
					"cannot setReadAccess for a user with null id");
		}
		setReadAccess(user.getObjectId(), allowed);
	}
	
	/**
	 * Set whether the given user is allowed to write this object.
	 */
	public void setWriteAccess(ParseUser user, boolean allowed) {
		if (user.getObjectId() == null) {
			throw new IllegalArgumentException(
					"cannot setWriteAccess for a user with null id");
		}
		setWriteAccess(user.getObjectId(), allowed);
	}
}
