package com.parse;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.UnsupportedEncodingException;
import java.util.ArrayList;
import java.util.Iterator;

import org.apache.http.HttpResponse;
import org.apache.http.HttpVersion;
import org.apache.http.client.ClientProtocolException;
import org.apache.http.client.HttpClient;
import org.apache.http.client.methods.HttpGet;
import org.apache.http.client.methods.HttpPost;
import org.apache.http.impl.client.DefaultHttpClient;
import org.apache.http.params.CoreProtocolPNames;
import org.json.JSONException;
import org.json.JSONObject;

import android.webkit.MimeTypeMap;

import com.parse.entity.mime.HttpMultipartMode;
import com.parse.entity.mime.MultipartEntity;
import com.parse.entity.mime.content.ByteArrayBody;
import com.parse.entity.mime.content.StringBody;

/**
 * 
 * ParseFile is a local representation of a file that is saved to the Parse cloud.
 * <p>
 * The workflow is to construct a ParseFile with data and optionally a filename. Then save it and set it as a field on a ParseObject.
 * <p>
 * Example:
 *	ParseFile file = new ParseFile("hello".getBytes());
 *	file.save();
 *	
 * ParseObject object = new ParseObject("TestObject");
 * object.put("file", file);
 * object.save();
 * 
 */
public class ParseFile {
	private String name = null;
	private String url = null;
	private byte[] data;
	private boolean isDirty = true;
	private boolean isDataAvailable = true;
	private boolean isCurrentlyFetching = false;
	
	private ArrayList<GetDataCallback> callbacks = new ArrayList<GetDataCallback>();
	
	/**
	 * Creates a new file from a byte array.
	 * 
	 * @param data The file's data.
	 */
	public ParseFile(byte[] data) {
		if (data.length > Parse.maxParseFileSize) {
			throw new IllegalArgumentException(String.format("ParseFile must be less than %i bytes", Parse.maxParseFileSize));
		}
		
		this.data = data;
	}
	
	/**
	 * Creates a new file from a byte array and a name. Giving a name with a proper file extension (e.g. ".png") is ideal
	 * because it allows Parse to deduce the content type of the file and set appropriate HTTP headers when it is fetched.
	 * 
	 * @param name The file's name, ideally with extension.
	 * @param data The file's data.
	 */
	public ParseFile(String name, byte[] data) {
		if (data.length > Parse.maxParseFileSize) {
			throw new IllegalArgumentException(String.format("ParseFile must be less than %i bytes", Parse.maxParseFileSize));
		}
		
		this.name = name;
		this.data = data;
	}
	
	ParseFile(String name, String url) {
		this.name = name;
		this.url = url;
		isDirty = false;
		isDataAvailable = false;
	}
	
	/**
	 * The filename. Before save is called, this is just the filename given by the user (if any). After save is called,
	 * that name gets prefixed with a unique identifier.
	 * 
	 * @return The file's name.
	 */
	public String getName() {
		return name;
	}
	
	/**
	 * Whether the file still needs to be saved.
	 * 
	 * @return Whether the file needs to be saved.
	 */
	boolean isDirty() {
		return isDirty;
	}
	
	/**
	 * Whether the file has available data.
	 */
	boolean isDataAvailable() {
		return isDataAvailable;
	}
	
	/**
	 * This returns the url of the file. It's only available after you save or after you get the file
	 * from a ParseObject.
	 * 
	 * @return The url of the file.
	 */
	public String getUrl() {
		return url;
	}
	
	// The lifecycle of creating a file is the following:
	// 1. We call the upload_file command on our servers with a requested name.
	// 2. We get back all the paramters required to make a POST directly to S3. At this point
	//    the object has all its metadata filled in.
	// 3. We make the S3 POST to upload the file.
	private ParseCommand constructFileUploadCommand() { 
		ParseCommand command = new ParseCommand("upload_file");
		
		if (name != null) {
			command.put("name", name);
		}
		
		return command;
	}
	
	private void handleFileUploadResult(JSONObject result) throws ParseException {		
		JSONObject postParams = null;

		try {
			name = result.getString("name");
			url = result.getString("url");
			postParams = result.getJSONObject("post_params");
		} catch (JSONException e) {
			throw new RuntimeException(e.getMessage());
		}
		
		String mimeType = null;
		MultipartEntity entity = new MultipartEntity(HttpMultipartMode.BROWSER_COMPATIBLE);
		
		if (name.lastIndexOf(".") != -1) {
			String extension = name.substring(name.lastIndexOf(".") + 1);
			mimeType = MimeTypeMap.getSingleton().getMimeTypeFromExtension(extension);
		} else {
			mimeType = "application/octet-stream";
		}
		
		try {
			entity.addPart("Content-Type", new StringBody(mimeType));
		} catch (UnsupportedEncodingException e) {
			throw new RuntimeException(e.getMessage());
		}
		
		@SuppressWarnings("unchecked")
		Iterator<String> keys = postParams.keys();
		while (keys.hasNext()) {
			String key = keys.next();
			
			try {
				entity.addPart(key, new StringBody(postParams.getString(key)));
			} catch (UnsupportedEncodingException e) {
				throw new RuntimeException(e.getMessage());
			} catch (JSONException e) {
				throw new RuntimeException(e.getMessage());
			}
		}
		
		entity.addPart("file", new ByteArrayBody(data, mimeType, "file"));
		
		HttpClient client = new DefaultHttpClient();		
	    client.getParams().setParameter(CoreProtocolPNames.PROTOCOL_VERSION, HttpVersion.HTTP_1_1);

		HttpPost post = null;
		try {
			post = new HttpPost(result.getString("post_url"));
		} catch (JSONException e) {
			throw new RuntimeException(e.getMessage());
		}
		
		post.setEntity(entity);
		
		try {
			HttpResponse response = client.execute(post);
								
			if (response.getStatusLine().getStatusCode() != 204) {
				throw new ParseException(ParseException.CONNECTION_FAILED, 
						String.format("Upload to S3 failed.", response.getStatusLine().getReasonPhrase()));
			}
			
			isDirty = false;
		} catch (ClientProtocolException e) {
			client.getConnectionManager().shutdown();
			throw new ParseException(ParseException.CONNECTION_FAILED, e.getMessage());
		} catch (IOException e) {
			client.getConnectionManager().shutdown();
			throw new ParseException(ParseException.CONNECTION_FAILED, e.getMessage());
		}
	}
	
	
	/**
	 * Saves the file to the Parse cloud synchronously.
	 */
	public void save() throws ParseException {
		if (!isDirty) { return; }
		
		ParseCommand command = constructFileUploadCommand();		
		JSONObject result = (JSONObject) command.perform();
		handleFileUploadResult(result);
	}
	
	/**
	 * Saves the file to the Parse cloud in a background thread.
	 * 
	 * @param callback A SaveCallback that gets called when the save completes.
	 */
	 public void saveInBackground(SaveCallback callback) {		
		if (!isDirty) { 
			callback.done(null); 
			return;
		}
				
		BackgroundTask<Void> saveTask = new BackgroundTask<Void>(callback) {
			public Void run() throws ParseException {
				ParseFile.this.save();
				return null;
			}
		};
		
		BackgroundTask.executeTask(saveTask);
	}
	
	public void saveInBackground() {
		saveInBackground(null);
	}
	
	/**
	 * Synchronously gets the data for this object. 
	 * You probably want to use {@link #getDataInBackground} instead unless you're
	 * already in a background thread.
	 */
	public byte[] getData() throws ParseException {
		if (isDataAvailable) {
			return data;
		}
		
		HttpClient client = new DefaultHttpClient();
		HttpGet get = new HttpGet(url);
		
		try {
			HttpResponse response = client.execute(get);
			InputStream responseStream = response.getEntity().getContent();
			
			data = convertStreamToByteArray(responseStream);
			isDataAvailable = true;
			
			return data;
		} catch (ClientProtocolException e) {
			throw new ParseException(ParseException.CONNECTION_FAILED, e.getMessage());
		} catch (IOException e) {
			throw new ParseException(ParseException.CONNECTION_FAILED, e.getMessage());
		}		
	}
	
	/**
	 * Gets the data for this object in a background thread.
	 * 
	 * @param callback A GetDataCallback that is called when the get completes.
	 */
	synchronized public void getDataInBackground(GetDataCallback callback) {
		if (isDataAvailable) {
			callback.done(data, null);
			return;
		}
		
		callbacks.add(callback);
		
		if (!isCurrentlyFetching) {
			isCurrentlyFetching = true;
			
			GetDataCallback myCallback = new GetDataCallback() {
				public void done(byte[] data, ParseException e) {
					synchronized (ParseFile.this) {						
						isCurrentlyFetching = false;
					}
					
					for (GetDataCallback callback : callbacks) {
						callback.done(data, e);
					}
					
					callbacks.clear();
				}
			};
			
			BackgroundTask<byte[]> getTask = new BackgroundTask<byte[]>(myCallback) {
				public byte[] run() throws ParseException {
					return ParseFile.this.getData();
				}
			};
			
			BackgroundTask.executeTask(getTask);
		}
	}

	// Converts an InputStream into byte[] -- useful for getting the result from an HttpGet.
	private static byte[] convertStreamToByteArray(InputStream is) throws IOException {
		ByteArrayOutputStream buffer = new ByteArrayOutputStream();
	
		int nRead;
		byte[] data = new byte[16384];
		
		while ((nRead = is.read(data, 0, data.length)) != -1) {
		  buffer.write(data, 0, nRead);
		}
	
		buffer.flush();

		return buffer.toByteArray();	
	}
	
	
}