package com.parse;

/**
 * 
 * ParseGeoPoint represents a latitude / longitude point that may be associated with
 * a key in a ParseObject or used as a reference point for geo queries.  This allows
 * proximity based queries on the key.
 * <p>
 * Only one key in a class may contain a GeoPoint.
 * <p> 
 * 
 * Example:
 *  ParseGeoPoint point = new ParseGeoPoint(30.0, -20.0);
 *  ParseObject object = new ParseObject("PlaceObject");
 *  object.put("location", point);
 *  object.save();
 *
 */

public class ParseGeoPoint {
	static double EARTH_MEAN_RADIUS_KM = 6371.0;
	static double EARTH_MEAN_RADIUS_MILE = 3958.8;

	private double latitude = 0.0;
	private double longitude = 0.0;
	
	/**
	 * Creates a new default point with latitude and longitude set to 0.0.
	 */
	public ParseGeoPoint() {}
	
	/**
	 * Creates a new point with the specified latitude and longitude.
	 * @param latitude The point's latitude.
	 * @param longitude The point's longitude.
	 */
	public ParseGeoPoint(double latitude, double longitude) {
		setLatitude(latitude);
		setLongitude(longitude);
	}
	
	/**
	 * Set latitude.  Valid range is (-90.0, 90.0).  Extremes should not be used.
	 * @param latitude The point's latitude. 
	 */
	public void setLatitude(double latitude) {
		if (latitude > 90.0 || latitude < -90.0) {
			throw new IllegalArgumentException("Latitude must be within the range (-90.0, 90.0).");
		}
		this.latitude = latitude;
	}
	
	/**
	 * Get latitude.
	 */
	public double getLatitude() {
		return latitude;
	}
	
	/**
	 * Set longitude.  Valid range is (-180.0, 180.0).  Extremes should not be used.
	 * @param longitude The point's longitude. 
	 */
	public void setLongitude(double longitude) {
		if (longitude > 180.0 || longitude < -180.0) {
			throw new IllegalArgumentException("Longitude must be within the range (-180.0, 180.0).");
		}
		this.longitude = longitude;
	}
	
	/**
	 * Get longitude.
	 */
	public double getLongitude() {
		return longitude;
	}
	
	/**
	 * Get distance in radians between this point and another GeoPoint.  This is the smallest angular
	 * distance between the two points.
	 * 
	 * @param point GeoPoint describing the other point being measured against.
	 */
	public double distanceInRadiansTo(ParseGeoPoint point) {
		double d2r = Math.PI / 180.0; // radian conversion factor
		double lat1rad = latitude * d2r;
		double long1rad = longitude * d2r;
		double lat2rad = point.getLatitude() * d2r;
		double long2rad = point.getLongitude() * d2r;
		double deltaLat = lat1rad - lat2rad;
		double deltaLong = long1rad - long2rad;
		double sinDeltaLatDiv2 = Math.sin(deltaLat / 2.);
		double sinDeltaLongDiv2 = Math.sin(deltaLong / 2.);
		// Square of half the straight line chord distance between both points. [0.0, 1.0]
		double a = sinDeltaLatDiv2 * sinDeltaLatDiv2
				+ Math.cos(lat1rad) * Math.cos(lat2rad) * sinDeltaLongDiv2 * sinDeltaLongDiv2;
		a = Math.min(1.0, a);
		return 2. * Math.asin(Math.sqrt(a));
	}
	
	/**
	 * Get distance between this point and another geopoint in kilometers.
	 * @param point GeoPoint describing the other point being measured against.
	 */
	public double distanceInKilometersTo(ParseGeoPoint point) {
		return distanceInRadiansTo(point) * EARTH_MEAN_RADIUS_KM;
	}

	/**
	 * Get distance between this point and another geopoint in kilometers.
	 * @param point GeoPoint describing the other point being measured against.
	 */
	public double distanceInMilesTo(ParseGeoPoint point) {
		return distanceInRadiansTo(point) * EARTH_MEAN_RADIUS_MILE;
	}	
}
