package com.parse;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.RandomAccessFile;
import java.io.UnsupportedEncodingException;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.SimpleTimeZone;
import java.util.UUID;

import org.json.JSONArray;
import org.json.JSONException;
import org.json.JSONObject;
import org.json.JSONTokener;

import android.content.Context;
import android.util.Log;

/**
 * The ParseObject is a local representation of data that can be saved and
 * retrieved from the Parse cloud.
 * <p>
 * The basic workflow for creating new data is to construct a new ParseObject,
 * use put() to fill it with data, and then use save() to persist to the database.
 * <p>
 * The basic workflow for accessing existing data is to use a {@link ParseQuery}
 * to specify which existing data to retrieve.
 */
public class ParseObject {
	private static final String TAG = "com.parse.ParseObject";
	static String server = "https://api.parse.com";
	static final String API_VERSION = "2";
	static final String VERSION_NAME = "0.4.14"; 
	static final String USER_CLASS_NAME = "_User";
	
	// This is the date format for createdAt and updatedAt.
	// It is only precise to the second.
	private static final DateFormat impreciseDateFormat;
	static {
		DateFormat format = new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss'Z'");
		format.setTimeZone(new SimpleTimeZone(0, "GMT"));
		impreciseDateFormat = format;
	}

	boolean dirty;
	
	private String objectId;
	private String className;
	
	private HashMap<String, ParsePointer> pointers;
	private HashMap<String, Object> data;
	private HashMap<String, ParseFieldOperation> operations;
	private HashMap<Object, ParseJSONCacheItem> hashedObjects;
	HashSet<String> dirtyKeys;
	
	private Date updatedAt;
	private Date createdAt;
	
	static String installationId;
	
	private Boolean isRunning = false;
	
	/**
	 * Constructs a new ParseObject with no data in it.
	 * A ParseObject constructed in this way will not have an objectId and will not
	 * persist to the database until save() is called.
	 * <p>
	 * Class names must be alphanumerical plus underscore, and start with a letter.
	 * It is recommended to name classes in CamelCaseLikeThis.
	 * @param theClassName The className for this ParseObject.
	 */
	public ParseObject(String theClassName) {
		pointers = new HashMap<String, ParsePointer>();
		data = new HashMap<String, Object>();
		operations = new HashMap<String, ParseFieldOperation>();
		hashedObjects = new HashMap<Object, ParseJSONCacheItem>();
		dirtyKeys = new HashSet<String>();
		className = theClassName;
		dirty = true;
	}
	
	/**
	 * Constructs a new ParseObject or ParseUser, depending on the class name. 
	 * @param theClassName
	 */
	static ParseObject build(String theClassName) {
		if (theClassName.equals(USER_CLASS_NAME)) {
			return new ParseUser();
		} else {
			return new ParseObject(theClassName);
		}
	}
	
    /**
     * Authenticates this client.
     * 
     * @deprecated
     * <code>ParseObject.initialize</code> has been replaced by
     * {@link Parse#initialize}, which also takes a <code>Context</code>.
	 */
	@Deprecated
	public static void initialize(String applicationId, String clientKey) {
		Parse.initialize(null, applicationId, clientKey);
	}
	
	static String getApplicationId() {
		Parse.checkInit();
		return Parse.applicationId;
	}
	
	// Used only to parse legacy created_at and updated_at from disk.
	private synchronized static Date impreciseParseDate(String encoded) {
		try {
			return impreciseDateFormat.parse(encoded);
		} catch (java.text.ParseException e) {
			// Should never happen
			Log.d(TAG, "could not parse date: " + encoded);
			return null;
		}
	}
	
	/**
	 * Loads a json object from disk given a string identifier.
	 * Returns an empty json object if there is no such file or it has a bad format.
	 */
	synchronized static JSONObject getDiskObject(Context context, String filename) {
		Parse.initializeContextIfNeeded(context);
		File file = new File(Parse.getParseDir(), filename);
		if (!file.exists()) {
			return null;
		}
		
		// Read the file
		String fileContent;
		try {
			RandomAccessFile f = new RandomAccessFile(file, "r");
			byte[] bytes = new byte[(int) f.length()];
			f.readFully(bytes);
			f.close();
			fileContent = new String(bytes, "UTF-8");
		} catch (IOException e) {
			return null;
		}
		
		// Convert to a JSONObject
		JSONTokener tokener = new JSONTokener(fileContent);
		try {
			return new JSONObject(tokener);
		} catch (JSONException e) {
			return null;
		}
	}
	
	/**
	 * Saves a json object to disk given a string identifier. Opposite of getDiskObject.
	 * Silences any errors.
	 */
	synchronized static void saveDiskObject(
			Context context, String filename, JSONObject object) {
		Parse.initializeContextIfNeeded(context);
		File file = new File(Parse.getParseDir(), filename);
		try {
			FileOutputStream out = new FileOutputStream(file);
			out.write(object.toString().getBytes("UTF-8"));
			out.close();
		} catch (UnsupportedEncodingException e) {
			return;
		} catch (IOException e) {
			return;
		}
	}
	
	protected void checkIfRunning() {
		checkIfRunning(false);
	}
	
	protected void checkIfRunning(boolean grabLock) {	
		synchronized (isRunning) {
			if (isRunning) {
				throw new RuntimeException("This query has an outstanding network connection. You have to wait until it's done.");
			} else if (grabLock) {
				isRunning = true;
			}
		}
	}
	
	protected void finishedRunning() {
		synchronized (isRunning) {
			isRunning = false;
		}
	}
	
	/**
	 * Saves the ParseObject to the a file on disk as JSON. 
	 * @param context The app context.
	 * @param filename The filename to save the object to.
	 */
	void saveToDisk(Context context, String filename) {
		if (isDirty()) {
			throw new RuntimeException("Can't serialize a dirty object to disk.");
		}
		
		JSONObject object = toJSONObject(true, true);
		saveDiskObject(context, filename, object);
	}
	
	void addToHashedObjects(Object object) {
		try {
			hashedObjects.put(object, new ParseJSONCacheItem(object));
		} catch (JSONException e) {
			throw new IllegalArgumentException("Couldn't serialize container value to JSON.");
		}
	}
	
	/**
	 * Retrieves a ParseObject from a file on disk.
	 * @param context The app context
	 * @param filename The filename to retrieve the object from.
	 * @return The ParseObject that was retrieved. If the file wasn't found, or the 
	 * 			contents of the file is an invalid ParseObject, returns null.
	 */
	static ParseObject getFromDisk(Context context, String filename) {
		JSONObject object = getDiskObject(context, filename);
		
		if (object == null) {
			return null;
		}
		
		try {
			ParseObject parseObject = build(object.getString("classname"));
			parseObject.mergeObject(object, false);
			return parseObject;
		} catch (JSONException e) {
			return null;
		}
	}
	
	/**
	 * Generates a unique identifier that is specific to this installation.
	 * @return A unique string that is also stored on disk.
	 */
	synchronized static String getInstallationId(Context context) {
		Parse.initializeContextIfNeeded(context);
		if (installationId == null) {
			File installationIdFile = new File(Parse.getParseDir(), "installationId");
			try {
				if (!installationIdFile.exists()) {
					installationId = UUID.randomUUID().toString();
					// Write the file
					FileOutputStream out = new FileOutputStream(installationIdFile);
					out.write(installationId.getBytes());
					out.close();
				} else {
					// Read the file
					RandomAccessFile f = new RandomAccessFile(installationIdFile, "r");
					byte[] bytes = new byte[(int) f.length()];
					f.readFully(bytes);
					f.close();
					installationId = new String(bytes);
				}
			} catch (IOException e) {
				// If we can't read the disk, use a UUID that's specific to
				// this invocation.
				if (installationId == null) {
					installationId = UUID.randomUUID().toString();
				}
			}
		}
		return installationId;
	}
	
	/**
	 * Accessor to the class name.
	 */
	public String getClassName() {
		return className;
	}
	
	/**
	 * This reports time as the server sees it, so that if you make changes to
	 * a ParseObject, then wait a while, and then call save(), the updated time
	 * will be the time of the save() call rather than the time the object was changed
	 * locally. 
	 * @return The last time this object was updated on the server.
	 */
	public Date getUpdatedAt() {
		return updatedAt;
	}
	
    /**
	 * Alias for <code>getUpdatedAt</code>.
	 */
	public Date updatedAt() {
		return getUpdatedAt();
	}
	
	/**
	 * This reports time as the server sees it, so that if you create a ParseObject, then
	 * wait a while, and then call save(), the creation time will be the time of the first
	 * save() call rather than the time the object was created locally.
	 * @return The first time this object was saved on the server.
	 */
	public Date getCreatedAt() {
		return createdAt;
	}
	
	/**
	 * Alias for <code>getCreatedAt</code>. 
	 */
	public Date createdAt() {
		return createdAt;
	}
	
	// Updates this object with data from the http api or disk json representation.
	// Sets dirtiness to false because it is assumed the server provided fresh data.
	void mergeObject(JSONObject object, boolean justCreated) {
		dirty = false;
		try {
			// The handlers for id, created_at, updated_at, and pointers are for
			// backward compatibility with old serialized users.
			if (object.has("id")) {
				if (objectId == null) {
					objectId = object.getString("id");
				}
			}
			if (object.has("created_at")) {
				String createdAtString = object.getString("created_at");
				if (createdAtString != null) {
					createdAt = impreciseParseDate(createdAtString);
				}
			}	
			if (object.has("updated_at")) {
				String updatedAtString = object.getString("updated_at");
				if (updatedAtString != null) {
					updatedAt = impreciseParseDate(updatedAtString);
				}
			}
			if (object.has("pointers")) {
				JSONObject newPointers = object.getJSONObject("pointers");
				@SuppressWarnings("unchecked")
				Iterator<String> keys = newPointers.keys();
				while (keys.hasNext()) {
					String key = keys.next();
					data.remove(key);
					JSONArray pointerArray = newPointers.getJSONArray(key);
					pointers.put(key, new ParsePointer(
							pointerArray.optString(0),
							pointerArray.optString(1)));
				}
			}
			
			if (object.has("data")) {
				JSONObject newData = object.getJSONObject("data");
				@SuppressWarnings("unchecked")
				Iterator<String> keys = newData.keys();
				while (keys.hasNext()) {
					String key = keys.next();

					if (key.equals("objectId")) {
						objectId = newData.getString(key);
						continue;
					}
					if (key.equals("createdAt")) {
						createdAt = Parse.parseDate(newData.getString(key));
						if (justCreated) {
							updatedAt = createdAt;
						}
						continue;
					}
					if (key.equals("updatedAt")) {
						updatedAt = Parse.parseDate(newData.getString(key));
						continue;
					}
					if (key.equals("ACL")) {
						setACL(ParseACL.createACLFromJSONObject(newData.getJSONObject(key)));
						continue;
					}
					if (key.equals("__type") || key.equals("className")) {
						continue;
					}
					
					data.remove(key);
					pointers.remove(key);
					Object value = newData.get(key);
					Object decodedObject = Parse.decodeJSONObject(value);
					
					if (decodedObject != null) {
						if (decodedObject instanceof ParsePointer) {
							pointers.put(key, (ParsePointer) decodedObject);
						} else {
							data.put(key, decodedObject);
						}
						
						if (Parse.isContainerObject(decodedObject)) {
							addToHashedObjects(decodedObject);
						}
					} else {
						data.put(key, value);
						
						if (Parse.isContainerObject(value)) {
							addToHashedObjects(value);
						}
					}
				}
			}
			
			if (object.has("dirty")) {
				dirty = object.getBoolean("dirty");
			}
		} catch (JSONException e) {
			throw new RuntimeException(e.getMessage());
		}
	}
	
	boolean hasDirtyChildren() {
		for (String key : data.keySet()) {
			Object object = data.get(key);
			if (object instanceof ParseObject && ((ParseObject)object).isDirty()) {
				return true;
			}
		}
		
		return false;
	}
	
	boolean hasChangedContainers() {
		for (String key : data.keySet()) {
			Object object = data.get(key);
			if (Parse.isContainerObject(object) && isKeyDirty(key)) {
				return true;
			}
		}
		
		return false;
	}
	
	boolean isDirty() {
		return this.isDirty(true);
	}
	
	boolean isDirty(boolean considerChildren) {
		return (dirty || !dirtyKeys.isEmpty() 
					  || hasChangedContainers() 
					  || (considerChildren && hasDirtyChildren())
			   );
	}
	
	boolean isKeyDirty(String key) {
		if (dirtyKeys.contains(key)) {
			return true;
		}
		
		Object object = data.get(key);
		
		if (object != null && Parse.isContainerObject(object)) {
			ParseJSONCacheItem oldCacheItem = hashedObjects.get(object);
			if (oldCacheItem == null) {
				dirtyKeys.add(key);
				return true;
			}
			
			ParseJSONCacheItem newCacheItem = null;
			
			try {
				newCacheItem = new ParseJSONCacheItem(object);
			} catch (JSONException e) {
				throw new RuntimeException();
			}
			
			if (newCacheItem == null || !oldCacheItem.equals(newCacheItem)) {
				hashedObjects.put(object, newCacheItem);
				dirtyKeys.add(key);
				return true;
			}
		}
		
		return false;
	}
	
	ParsePointer address() {
		if (objectId == null) {
			return null;
		}
		return new ParsePointer(className, objectId);
	}

	/**
	 * Accessor to the object id.
	 * An object id is assigned as soon as an object is saved to the server.
	 * The combination of a className and an objectId uniquely identifies an
	 * object in your application.
	 * @return The object id.
	 */
	public String getObjectId() {
		return objectId;
	}
	
	/**
	 * Setter for the object id.
	 * In general you do not need to use this.
	 * However, in some cases this can be convenient. For example, if you are serializing
	 * a ParseObject yourself and wish to recreate it, you can use this to recreate the
	 * ParseObject exactly.
	 */
	public void setObjectId(String newObjectId) {
		checkIfRunning();
		dirty = true;
		objectId = newObjectId;
	}
	
	/**
	 * Alias for <code>getObjectId</code>.
	 */
	public String objectId() {
		return getObjectId();
	}
	
	ParseCommand constructSaveCommand() throws ParseException {
		// Recursively save children
		for (Object value : data.values()) {
			if (value instanceof ParseObject) {
				((ParseObject) value).save();
			}
		}
		
		if (!isDirty()) {
			return null;
		}
		
		// Get the JSON representation of the object, and use some of the information
		// to construct the command
		JSONObject objectJSON = toJSONObject(false, false);
		
		// Construct the command
		String op = objectId == null ? "create" : "update";
		ParseCommand command = new ParseCommand(op);
		command.enableRetrying();
		
		command.put("classname", className);
				
		try {
			command.put("data", objectJSON.getJSONObject("data"));
		} catch (JSONException e) {
			throw new RuntimeException("could not decode data");
		}
		
		return command;
	}
	
	/**
	 * Converts a ParseObject to its JSON representation: 
	 * 
	 *  {
	 * 		data: {
	 * 			// data fields, including objectId, createdAt, updatedAt
	 * 		},
	 * 		classname: class name for the object
	 * 		dirty: boolean state of whether the object is dirty
	 * 		operations: { } // operations per field
	 *  }
	 *  
	 *  If includeDates is false, createdAt and updatedAt are not included.
	 *  
	 *  If allKeys is true, we send up all keys regardless of whether they are dirty. If allKeys is false, we only send the dirty keys.
	 */
	JSONObject toJSONObject(boolean includeDates, boolean allKeys) {
		// Public data goes in dataJSON; special fields go in objectJSON.
		JSONObject objectJSON = new JSONObject();
		JSONObject dataJSON = new JSONObject();

		try {		
			// Serialize the data
			for (String key : data.keySet()) {
				if (!allKeys && !isKeyDirty(key)) {
					continue;
				}
				
				Object object = data.get(key);
				
				// Pull it out of the json cache if it's there.
				if (Parse.isContainerObject(object) && hashedObjects.containsKey(object)) {
					dataJSON.put(key, hashedObjects.get(object).getJSONObject());
				} else {
					dataJSON.put(key, Parse.maybeReferenceAndEncode(object));
				}
			}
			
			// Serialize the pointers
			for (String key : pointers.keySet()) {
				ParsePointer pointer = pointers.get(key);
				dataJSON.put(key, Parse.pointerToJSONObject(pointer));
			}
			
			// Serialize the operations
			for (String key : operations.keySet()) {
				ParseFieldOperation operation = operations.get(key);
				dataJSON.put(key, Parse.maybeEncodeJSONObject(operation));
			}

			if (includeDates) {
				if (createdAt != null) {
					dataJSON.put("createdAt", Parse.encodeDate(createdAt));
				}
				if (updatedAt != null) {
					dataJSON.put("updatedAt", Parse.encodeDate(updatedAt));
				}
			}

			if (objectId != null) {
				dataJSON.put("objectId", objectId);
			}

			objectJSON.put("data", dataJSON);			
			objectJSON.put("classname", className);
		} catch (JSONException e) {
			throw new RuntimeException("could not serialize object to JSON");
		}
		
		return objectJSON;
	}

	protected void handleSaveResult(String op, JSONObject result) {
		boolean justCreated = op.equals("create") || op.equals("user_signup"); 
		mergeObject(result, justCreated);
		dirtyKeys = new HashSet<String>();
		operations = new HashMap<String, ParseFieldOperation>();
		dirty = false;
	}
	
	void validateSave() {
		return;
	}
	
	/**
	 * Saves this object to the server.
	 * Typically, you should use {@link #saveInBackground} instead of this,
	 * unless you are managing your own threading.
	 * 
	 * @throws ParseException Throws an exception if the server is inaccessible.
	 */
	public void save() throws ParseException {
		save(true);
	}
	
	protected void save(boolean needsLock) throws ParseException {
		if (needsLock) { checkIfRunning(true); }

		try {
			validateSave();
			
			ParseCommand command = constructSaveCommand();
			if (command == null) {
				return;
			}
			
			JSONObject result = (JSONObject) command.perform();
			handleSaveResult(command.op, result);
		} finally {
			finishedRunning();
		}
	}
	
	/**
	 * Saves this object to the server in a background thread.
	 * This is preferable to using save(), unless your code is
	 * already running from a background thread.
	 * @param callback callback.done(e) is called when the save completes.
	 */
	public void saveInBackground(SaveCallback callback) {
		checkIfRunning(true);
		
		BackgroundTask<Void> saveTask = new BackgroundTask<Void>(callback) {
			public Void run() throws ParseException {
				ParseObject.this.save(false);
				return null;
			}
		};
		BackgroundTask.executeTask(saveTask);
	}
	
	/**
	 * Saves this object to the server in a background thread.
	 * Use this when you do not have code to run on completion of the push.
	 */
	public void saveInBackground() {
		saveInBackground(null);
	}
	
	/**
	 * Refreshes this object with the data from the server. Call this whenever you want
	 * the state of the object to reflect exactly what is on the server.
	 * @throws ParseException Throws an exception if the server is inaccessible.
	 */
	public void refresh() throws ParseException {
		refresh(true);
	}
	
	protected void refresh(boolean needsLock) throws ParseException {
		if (needsLock) { checkIfRunning(true); }
		
		try {
			if (objectId == null) {
				throw new IllegalArgumentException("Cannot refresh an object that hasn't been saved to the server.");
			}
			
			ParseCommand command = new ParseCommand("get");
			command.enableRetrying();
			command.put("classname", className);
			JSONObject data = new JSONObject();
			try {
				data.put("objectId", objectId);
			} catch (JSONException e) {
				throw new RuntimeException(e.getMessage());
			}
			command.put("data", data);
			JSONObject result = (JSONObject) command.perform();
			
			mergeObject(result, false);
		} finally {
			finishedRunning();
		}
	}
	
	/**
	 * Refreshes this object with the data from the server in a background thread.
	 * This is preferable to using refresh(), unless your code is
	 * already running from a background thread.
	 * @param callback callback.done(object, e) is called when the refresh completes.
	 */
	public void refreshInBackground(RefreshCallback callback) {
		checkIfRunning(true);
		
		BackgroundTask<ParseObject> refreshTask = new BackgroundTask<ParseObject>(callback) {
			public ParseObject run() throws ParseException {
				ParseObject.this.refresh(false);
				return ParseObject.this;
			}
		};
		BackgroundTask.executeTask(refreshTask);
	}
	
	ParseCommand constructDeleteCommand() throws ParseException {
		ParseCommand command = new ParseCommand("delete");
		command.enableRetrying();
		command.put("classname", className);
		JSONObject data = new JSONObject();
		try {
			data.put("objectId", objectId);
		} catch (JSONException e) {
			throw new RuntimeException(e.getMessage());
		}
		command.put("data", data);
		return command;
	}
	
	// Validates the delete method
	void validateDelete() {
		return;
	}
	
	/**
	 * Deletes this object on the server.
	 * This does not delete or destroy the object locally.
	 * @throws ParseException Throws an error if the object does not exist or if the internet fails.
	 */
	public void delete() throws ParseException {
		delete(true);
	}
	
	protected void delete(boolean needsLock) throws ParseException {
		if (needsLock) { checkIfRunning(true); }
		
		try {
			validateDelete();
			
			if (objectId == null) {
				return;
			}
			
			ParseCommand command = constructDeleteCommand();
			command.perform();
			dirty = true;
		} finally {
			finishedRunning();
		}
	}
	
	/**
	 * Deletes this object on the server in a background thread.
	 * This is preferable to using delete(), unless your code is already
	 * running from a background thread.
	 * @param callback callback.done(e) is called when the save completes.
	 */
	public void deleteInBackground(DeleteCallback callback) {
		checkIfRunning(true);
		validateDelete();
		
		BackgroundTask<Void> deleteTask = new BackgroundTask<Void>(callback) {
			public Void run() throws ParseException {
				ParseObject.this.delete(false);
				return null;
			}
		};
		BackgroundTask.executeTask(deleteTask);
	}
	
	/**
	 * Deletes this object on the server in a background thread.
	 * Does nothing in particular when the save completes.
	 * Use this when you don't care if the delete works.
	 */
	public void deleteInBackground() {
		deleteInBackground(null);
	}
	
	/**
	 * Saves each object in the provided list.
	 * This is faster than saving each object individually because it batches the requests.
	 * @param objects The objects to save.
	 * @throws ParseException Throws an exception if the server returns an error or is inaccessible.
	 */
	public static void saveAll(List<ParseObject> objects) throws ParseException {
		// Package all save commands together
		JSONArray dirtyCommands = new JSONArray();
		List<ParseObject> dirtyObjects = new ArrayList<ParseObject>();
		List<String> dirtyOps = new ArrayList<String>();
		for (ParseObject object : objects) {
			object.validateSave();
			ParseCommand command = object.constructSaveCommand();
			if (command != null) {
				JSONObject jsonCommand = command.toJSONObject();
				dirtyCommands.put(jsonCommand);
				dirtyObjects.add(object);
				dirtyOps.add(command.op);
			}
		}
		
		// Send the multi request
		ParseCommand multiCommand = new ParseCommand("multi");
		multiCommand.put("commands", dirtyCommands);
		JSONArray results = (JSONArray) multiCommand.perform();

		// Handle the responses
		for (int i = 0; i < dirtyObjects.size(); ++i) {
			String op = dirtyOps.get(i);
			try {
				JSONObject result = results.getJSONObject(i);
				dirtyObjects.get(i).handleSaveResult(op, result);
			} catch (JSONException e) {
				throw new RuntimeException(e.getMessage());
			}
		}
	}
	
	/**
	 * Saves each object in the provided list to the server in a background thread.
	 * This is preferable to using saveAll, unless your code is already running
	 * from a background thread.
	 * @param objects The objects to save.
	 * @param callback callback.done(e) is called when the save completes.
	 */
	public static void saveAllInBackground(List<ParseObject> objects, SaveCallback callback) {
		final List<ParseObject> finalObjects = objects;
		BackgroundTask<Void> saveTask = new BackgroundTask<Void>(callback) {
			public Void run() throws ParseException {
				ParseObject.saveAll(finalObjects);
				return null;
			}
		};
		BackgroundTask.executeTask(saveTask);
	}
	
	/**
	 * Saves each object to the server in a background thread.
	 * Does nothing in particular when the save completes.
	 * Use this when you don't care if the save works.
	 * @param objects The objects to save.
	 */
	public static void saveAllInBackground(List<ParseObject> objects) {
		saveAllInBackground(objects, null);
	}
	
	/**
	 * Removes a key from this object's data if it exists.
	 * @param key The key to remove.
	 */
	public void remove(String key) {
		checkIfRunning();
		Object object = data.get(key);
		ParsePointer pointer = pointers.get(key);
		
		if (object != null || pointer != null) {
			dirtyKeys.add(key);			
			operations.put(key, new ParseDeleteOperation());
		}
		
		if (object != null) {
			hashedObjects.remove(object);
			data.remove(key);
		} else if (pointer != null) {
			pointers.remove(key);
		}		
	}
	
	/**
	 * Whether this object has a particular key. Same as containsKey.
	 * @param key The key to check for
	 * @return Returns whether this object contains the key
	 */
	public boolean has(String key) {
		return containsKey(key);
	}
	
	/**
	 * Add a key-value pair to this object.
	 * It is recommended to name keys in partialCamelCaseLikeThis.
	 * @param key Keys must be alphanumerical plus underscore, and start with a letter.
	 * @param value Values may be numerical, String, JSONObject, JSONArray, JSONObject.NULL, 
	 * or other ParseObjects. value may not be <code>null</code>.
	 */
	public void put(String key, Object value) {
		checkIfRunning();
		if (key == null) {
			throw new IllegalArgumentException("key may not be null.");
		}
		
		if (value == null) {
			throw new IllegalArgumentException("value may not be null.");
		}
		
		if (value instanceof ParseFile && ((ParseFile) value).isDirty()) {
			throw new IllegalArgumentException("ParseFile must be saved before being set on a ParseObject.");
		}
		
		if (!(value instanceof JSONObject ||
			  value instanceof JSONArray ||
			  value instanceof String ||
			  value instanceof Number ||
			  value instanceof Boolean ||
			  value == JSONObject.NULL ||
			  value instanceof ParseObject ||
			  value instanceof ParseACL ||
			  value instanceof ParseFile ||
			  value instanceof ParseGeoPoint ||
			  value instanceof Date ||
			  value instanceof byte[])) {
			throw new IllegalArgumentException(
					"invalid type for value: " + value.getClass().toString());
		}
		
		// If we have a pending delete for this key, let's cancel that.
		if (operations.containsKey(key) && operations.get(key) instanceof ParseDeleteOperation) {
			operations.remove(key);
		}
		
		// If this is a container type, let's hash it.
		if (Parse.isContainerObject(value)) {
			addToHashedObjects(value);
		}
				
		data.put(key, value);
		pointers.remove(key);
		dirtyKeys.add(key);
	}

	/**
	 * Atomically increments the given key by 1.
	 * @param key The key to increment.
	 */
	public void increment(String key) {
		increment(key, 1);
	}
	
	/**
	 * Atomically increments the given key by the given number.
	 * @param key The key to increment.
	 * @param amount The amount to increment by.
	 */
	public void increment(String key, Number amount) {
		checkIfRunning();
		if (data.containsKey(key) && !(data.get(key) instanceof Number)) {
			throw new IllegalArgumentException("Can't increment a key that isn't a number.");
		} 
		
		if (objectId == null) {
			if (data.containsKey(key)) {
				this.put(key, Parse.addNumbers(this.getNumber(key), amount));
			} else {
				this.put(key, amount);
			}
			return;
		}
		
		ParseFieldOperation operation = operations.get(key);
		ParseIncrementOperation incrementOperation;
		
		if (operation != null) {
			if (!(operation instanceof ParseIncrementOperation)) {
				throw new RuntimeException("Can't increment a key that has another pending operation (likely deletion).");
			}
			
			incrementOperation = (ParseIncrementOperation)operation;
			incrementOperation.setAmount(Parse.addNumbers(incrementOperation.getAmount(), amount));
		} else {
			incrementOperation = new ParseIncrementOperation();
			incrementOperation.setAmount(amount);
		}
		
		if (incrementOperation.getAmount().equals(0)) {
			operations.remove(key);
			dirtyKeys.remove(key);
		} else {
			operations.put(key, incrementOperation);
			dirtyKeys.add(key);			
		}
	}
	
	/**
	 * Whether this object has a particular key. Same as 'has'.
	 * @param key The key to check for
	 * @return Returns whether this object contains the key
	 */
	public boolean containsKey(String key) {
		return data.containsKey(key) || pointers.containsKey(key);
	}
	
	/**
	 * Access a string value.
	 * @param key The key to access the value for.
	 * @return Returns null if there is no such key or if it is not a String.
	 */
	public String getString(String key) {
		if (!data.containsKey(key)) {
			return null;
		}
		Object value = data.get(key);
		if (!(value instanceof String)) {
			return null;
		}
		return (String) value;
	}
	
	
	
	/**
	 * Access a byte array value.
	 * @param key The key to access the value for.
	 * @return Returns null if there is no such key or if it is not a byte array.
	 */
	public byte[] getBytes(String key) {
		if (!data.containsKey(key)) {
			return null;
		}
		Object value = data.get(key);
		if (!(value instanceof byte[])) {
			return null;
		}
		
		return (byte[]) value;
	}
	
	/**
	 * Access a numerical value.
	 * @param key The key to access the value for.
	 * @return Returns null if there is no such key or if it is not a Number.
	 */
	public Number getNumber(String key) {
		if (!data.containsKey(key)) {
			return null;
		}
		Object value = data.get(key);
		if (!(value instanceof Number)) {
			return null;
		}
		return (Number) value;
	}
	
	/**
	 * Access a JSONArray value.
	 * @param key The key to access the value for.
	 * @return Returns null if there is no such key or if it is not a JSONArray.
	 */
	public JSONArray getJSONArray(String key) {
		if (!data.containsKey(key)) {
			return null;
		}
		Object value = data.get(key);
		if (!(value instanceof JSONArray)) {
			return null;
		}
		return (JSONArray) value;
	}
	
	/**
	 * Access a JSONObject value.
	 * @param key The key to access the value for.
	 * @return Returns null if there is no such key or if it is not a JSONObject.
	 */
	public JSONObject getJSONObject(String key) {
		if (!data.containsKey(key)) {
			return null;
		}
		Object value = data.get(key);
		if (!(value instanceof JSONObject)) {
			return null;
		}
		return (JSONObject) value;
	}
	
	/**
	 * Access an int value.
	 * @param key The key to access the value for.
	 * @return Returns 0 if there is no such key or if it is not a JSONObject.
	 */
	public int getInt(String key) {
		Number number = getNumber(key);
		if (number == null) {
			return 0;
		}
		return number.intValue();
	}
	
	/**
	 * Access a double value.
	 * @param key The key to access the value for.
	 * @return Returns 0 if there is no such key or if it is not a double.
	 */
	public double getDouble(String key) {
		Number number = getNumber(key);
		if (number == null) {
			return 0;
		}
		return number.doubleValue();
	}
	
	/**
	 * Access a long value.
	 * @param key The key to access the value for.
	 * @return Returns null if there is no such key or if it is not a long.
	 */
	public long getLong(String key) {
		Number number = getNumber(key);
		if (number == null) {
			return 0;
		}
		return number.longValue();
	}
	
	/**
	 * Access a boolean value.
	 * @param key The key to access the value for.
	 * @return Returns false if there is no such key or if it is not a boolean.
	 */
	public boolean getBoolean(String key) {
		if (!data.containsKey(key)) {
			return false;
		}
		Object value = data.get(key);
		if (!(value instanceof Boolean)) {
			return false;
		}
		return ((Boolean) value).booleanValue();
	}
	
	/**
	 * Access a ParseObject value.
	 * @param key The key to access the value for.
	 * @return Returns null if there is no such key or if it is not a ParseObject.
	 * @throws Throws a ParseException if the server could not be reached to load a
	 * relational value.
	 */
	public ParseObject getParseObject(String key) throws ParseException {
		Object value = get(key);
		if (!(value instanceof ParseObject)) {
			return null;
		}
		return (ParseObject) value;
	}
	
	/**
	 * Access a ParseUser value.
	 * @param key The key to access the value for.
	 * @return Returns null if there is no such key or if it is not a ParseUser.
	 * @throws Throws a ParseException if the server could not be reached to load a
	 * relational value.
	 */
	public ParseUser getParseUser(String key) throws ParseException {
		Object value = get(key);
		if (!(value instanceof ParseUser)) {
			return null;
		}
		return (ParseUser) value;
	}
	
	/**
	 * Access a ParseGeoPoint value.
	 * @param key The key to access the value for
	 * @return Returns null if there is no such key or if it is not a ParseGeoPoint.
	 * @throws Throws a ParseException if the server could not be reached to load a
	 * relational value.
	 */
	public ParseGeoPoint getParseGeoPoint(String key) throws ParseException {
		Object value = get(key);
		if (!(value instanceof ParseGeoPoint)) {
			return null;
		}
		return (ParseGeoPoint) value;
	}
	
	/**
	 * Access the ParseACL governing this object.
	 */
	public ParseACL getACL() {
		Object acl = data.get("ACL");
		if (acl == null) {
			return null;
		}
		if (!(acl instanceof ParseACL)) {
			throw new RuntimeException("only ACLs can be stored in the ACL key");
		}
		return (ParseACL) acl;
	}
	
	/**
	 * Set the ParseACL governing this object.
	 */
	public void setACL(ParseACL acl) {
		put("ACL", acl);
	}
	
	/**
	 * Access a value.
	 * In most cases it is more convenient to use a helper function such as getString
	 * or getInt.
	 * @param key The key to access the value for.
	 * @return Returns null if there is no such key.
	 * @throws Throws a ParseException if the server could not be reached to load a
	 * relational value.
	 */
	public Object get(String key) throws ParseException {
		if (data.containsKey(key)) {
			return data.get(key);
		}
		
		ParsePointer pointer = pointers.get(key);
		if (pointer == null) {
			return null;
		}

		ParseQuery query = new ParseQuery(pointer.className);
		ParseObject child = query.get(pointer.objectId);
		data.put(key, child);
		pointers.remove(key);
		return child;
	}
	
	/**
	 * Access a ParseObject value and fetch it in the background if necessary.
	 * @param key The key to access the value for.
	 * @param callback callback.done(object, e) will be called when the fetch completes.
	 */
	public void getParseObjectInBackground(String key, GetCallback callback) {
		final String finalKey = key;
		BackgroundTask<ParseObject> getTask = new BackgroundTask<ParseObject>(callback) {
			public ParseObject run() throws ParseException {
				return ParseObject.this.getParseObject(finalKey);
			}
		};
		BackgroundTask.executeTask(getTask);
	}
}
