package com.parse;

import org.json.JSONException;
import org.json.JSONObject;

import android.util.Log;

/**
 * The ParsePush is a local representation of data that can be sent as a push notification.
 * <p>
 * The typical workflow for sending a push notification from the client is to construct
 * a new ParsePush, use the setter functions to fill it with data, and then use
 * {@link #sendInBackground()} to send it.
 */
public class ParsePush {
	private static final String TAG = "com.parse.ParsePush";
	private String mChannel = "";
	private boolean mPushToIOS = false;
	private boolean mPushToAndroid = true;
	private JSONObject mData;
	
	/**
	 * Creates a new push notification to send it.
	 * The default channel is the empty string, also known as the global broadcast
	 * channel.
	 * Before sending the push notification you must call either {@link #setMessage(String)}
	 * or {@link #setData(JSONObject)}.
	 */
	public ParsePush() {}
	
	/** Set the channel this push notification will be sent to.
	 */
	public void setChannel(String channel) {
		if (channel == null) {
			throw new IllegalArgumentException("channel cannot be null");
		}
		mChannel = channel;
	}
	
	/**
	 * Set whether this push notification will go to iOS devices.
	 *
	 * Defaults to false. Before using this, you must set up iOS push certificates.
	 */
	public void setPushToIOS(boolean pushToIOS) {
		mPushToIOS = pushToIOS;
	}
	
	/**
	 * Set whether this push notification will go to Android devices.
	 * 
	 * Defaults to true.
	 */
	public void setPushToAndroid(boolean pushToAndroid) {
		mPushToAndroid = pushToAndroid;
	}
	
	/**
	 * Sets the entire data of the push message.
	 * Typically you should use {@link #setMessage(String)} instead. 
	 * 
	 * See the push guide for more details.
	 */
	public void setData(JSONObject data) {
		mData = data;
	}
	
	/**
	 * Sets the message that will be shown in the notification.
	 */
	public void setMessage(String message) {
		JSONObject data = new JSONObject();
		try {
			data.put("alert", message);
		} catch (JSONException e) {
			Log.d(TAG, "JSONException in setMessage: " + e.getMessage());
		}
		setData(data);
	}

	/**
	 * Sends this push notification while blocking this thread until the push
	 * notification has successfully reached the Parse servers.
	 * Typically, you should use {@link #sendInBackground()} instead of this,
	 * unless you are managing your own threading.
	 * 
	 * @throws ParseException Throws an exception if the server is inaccessible.
	 */
	public void send() throws ParseException {
		ParseCommand command = new ParseCommand("client_push");
		if (mData == null) {
			throw new IllegalArgumentException(
					"cannot send a push without calling either setMessage or setData");
		}
		command.put("data", mData);
		command.put("channel", mChannel);
		if (mPushToIOS) {
			if (mPushToAndroid) {
				// Push to both by not including a 'type' parameter
			} else {
				command.put("type", "ios");
			}
		} else if (mPushToAndroid) {
			command.put("type", "android");
		} else {
			throw new IllegalArgumentException(
					"cannot push if both pushToIOS and pushToAndroid are false");
		}
		
		command.perform();
	}
	
	/**
	 * Sends this push notification in a background thread.
	 * This is preferable to using <code>send()</code>, unless your code is
	 * already running from a background thread.
	 * @param callback callback.done(e) is called when the send completes.
	 */
	public void sendInBackground(SendCallback callback) {
		BackgroundTask<Void> sendTask = new BackgroundTask<Void>(callback) {
			public Void run() throws ParseException {
				ParsePush.this.send();
				return null;
			}
		};
		BackgroundTask.executeTask(sendTask);
	}
	
	/**
	 * Sends this push notification in a background thread.
	 * Use this when you do not have code to run on completion of the push.
	 */
	public void sendInBackground() {
		sendInBackground(null);
	}
}
