package com.parse;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.regex.Pattern;

import org.json.JSONArray;
import org.json.JSONException;
import org.json.JSONObject;

import android.util.Log;

/**
 * The ParseQuery class defines a query that is used to fetch ParseObjects. The most
 * common use case is finding all objects that match a query through the
 * {@link #findInBackground findInBackground} method, using a {@link FindCallback}.
 * 
 * For example, this sample code fetches all objects of class <code>"MyClass"</code>.
 * It calls a different function depending on whether the fetch succeeded or not.
 * 
 * <pre>
 * ParseQuery query = new ParseQuery("MyClass");
 * query.findInBackground(new FindCallback() {
 *     public void done(List&lt;ParseObject&gt; objects, ParseException e) {
 *         if (e == null) {
 *             objectsWereRetrievedSuccessfully(objects);
 *         } else {
 *             objectRetrievalFailed();
 *         }
 *     }
 * }
 * </pre>
 * 
 * A ParseQuery can also be used to retrieve a single object whose id is known, through
 * the {@link #getInBackground getInBackground} method, using a {@link GetCallback}.
 * 
 * For example, this sample code fetches an object of class <code>"MyClass"</code> and
 * id <code>myId</code>.
 * It calls a different function depending on whether the fetch succeeded or not.
 * 
 * <pre>
 * ParseQuery query = new ParseQuery("MyClass");
 * ParseObject.getInBackground(myId, new GetCallback() {
 *     public void done(ParseObject object, ParseException e) {
 *         if (e == null) {
 *             objectWasRetrievedSuccessfully(object);
 *         } else {
 *             objectRetrievalFailed();
 *         }
 *     }
 * }
 * </pre>
 * 
 * A ParseQuery can also be used to count the number of objects that match the query
 * without retrieving all of those objects.
 * 
 * For example, this sample code counts the number of objects of the class
 * <code>"MyClass"</code>.
 * 
 * <pre>
 * ParseQuery query = new ParseQuery("MyClass");
 * query.countInBackground(new FindCallback() {
 *     public void done(int count, ParseException e) {
 *         if (e == null) {
 *             objectsWereCounted(count);
 *         } else {
 *             objectCountFailed();
 *         }
 *     }
 * }
 * </pre>
 * 
 * Using the callback methods is usually preferred because the network operation will not
 * block the calling thread. However, in some cases it may be easier to use the
 * {@link #find find}, {@link #get get} or {@link #count count} calls, which do block the
 * calling thread. For example, if your application has already
 * spawned a background task to perform work, that background task could use the blocking
 * calls and avoid the code complexity of callbacks.
 */
public class ParseQuery {
	private static final String TAG = "com.parse.ParseQuery";
	
	private String className;
	private HashMap<String, Object> where;
	private ArrayList<String> include;
	private int limit;
	private int skip;
	private String order;
	
	private BackgroundTask<?> currentTask = null;
	private ParseCommand currentCommand = null;
	private Boolean isRunning = false;
	
	public enum CachePolicy {
		IGNORE_CACHE,
		CACHE_ONLY,
		NETWORK_ONLY,
		CACHE_ELSE_NETWORK,
		NETWORK_ELSE_CACHE,
		CACHE_THEN_NETWORK
	}
	
	private CachePolicy cachePolicy;
	
	/**
	 * Constructs a query.
	 * A default query with no further parameters will retrieve all ParseObjects
	 * of the provided class.
	 * @param theClassName The name of the class to retrieve ParseObjects for.
	 */
	public ParseQuery(String theClassName) {
		className = theClassName;
		limit = -1; // negative limits mean, do not send a limit
		skip = 0;
		where = new HashMap<String, Object>();
		include = new ArrayList<String>();
		cachePolicy = CachePolicy.IGNORE_CACHE;
	}
	
	/**
	 * Constructs a query for ParseUsers
	 */
	public static ParseQuery getUserQuery() {
		return new ParseQuery(ParseObject.USER_CLASS_NAME);
	}
	
	private void checkIfRunning() {
		checkIfRunning(false);
	}
	
	private void checkIfRunning(boolean grabLock) {	
		synchronized (isRunning) {
			if (isRunning) {
				throw new RuntimeException("This query has an outstanding network connection. You have to wait until it's done.");
			} else if (grabLock) {
				isRunning = true;
			}
		}
	}
	
	private void finishedRunning() {
		synchronized (isRunning) {
			isRunning = false;
			currentTask = null;
			currentCommand = null;
		}
	}
		
	// Creates a ParseCommand for a 'find' operation.
	// If the query is invalid, returns null.
	private ParseCommand makeFindCommand() {
		ParseCommand command = new ParseCommand("find");
		command.enableRetrying();
		command.put("classname", className);
		
		try {
			JSONObject whereData = new JSONObject();
			for (String key : where.keySet()) {
				Object value = where.get(key);
				whereData.put(key, Parse.maybeReferenceAndEncode(value));				
			}
			
			command.put("data", whereData);
			if (limit >= 0) {
				command.put("limit", limit);
			}
			if (skip > 0) {
				command.put("skip", skip);
			}
			if (order != null) {
				command.put("order", order);
			}
			if (!include.isEmpty()) {
				command.put("include", Parse.join(include, ","));
			}
		} catch (JSONException e) {
			throw new RuntimeException(e.getMessage());
		}
		
		return command;
	}
	
	// Converts the JSONArray that represents the results of a find command to
	// an ArrayList<ParseObject>.
	private ArrayList<ParseObject> convertFindResponse(JSONObject response) throws JSONException {
		ArrayList<ParseObject> answer = new ArrayList<ParseObject>();
		JSONArray results = response.getJSONArray("results");
		if (results == null) {
			Log.d(TAG, "null results in find response");
		} else {
			for (int i = 0; i < results.length(); ++i) {
				ParseObject object = ParseObject.build(className);
				object.mergeObject(results.getJSONObject(i), false);
				answer.add(object);
			}
		}
		return answer;
	}

	private ArrayList<ParseObject> findWithCachePolicy(CachePolicy policy) throws ParseException {		
		switch(policy) {
		case IGNORE_CACHE:
		case NETWORK_ONLY:
			return findFromNetwork();
		case CACHE_ONLY:
			return findFromCache();
		case CACHE_ELSE_NETWORK:
			try {
				return findFromCache();
			} catch (ParseException e) {
				return findFromNetwork();
			}
		case NETWORK_ELSE_CACHE:
			try {
				return findFromNetwork();
			} catch (ParseException e) {
				if (e.getCode() != ParseException.CONNECTION_FAILED) {
					// This is an error with the query, not with the network
					throw e;
				}
				return findFromCache();
			}
		case CACHE_THEN_NETWORK:
			throw new RuntimeException("You cannot use the cache policy CACHE_THEN_NETWORK with find()");
		default:
			throw new RuntimeException("Unknown cache policy: " + cachePolicy);
		}
	}
	
	/**
	 * Cancels the current network request (if one is running).
	 */
	public void cancel() {
		if (currentTask != null) {
			currentTask.cancel(true);
			currentTask = null;
		}
		
		if (currentCommand != null) {
			currentCommand.cancel();
			currentCommand = null;
		}
		
		isRunning = false;
	}
	
	/** Retrieves a list of ParseObjects that satisfy this query.
	 * Uses the network and/or the cache, depending on the cache policy.
	 * @return A list of all ParseObjects obeying the conditions set in this query.
	 * @throws ParseException
	 */
	public ArrayList<ParseObject> find() throws ParseException {		
		checkIfRunning(true);
		ArrayList<ParseObject> result;
		try {
			result = findWithCachePolicy(cachePolicy);			
		} finally {
			finishedRunning();
		}
		return result;
	}
	
	/** Retrieves a list of ParseObjects that satisfy this query from the server.
	 * @return A list of all ParseObjects obeying the conditions set in this query.
	 * @throws ParseException
	 */
	private ArrayList<ParseObject> findFromNetwork() throws ParseException {
		currentCommand = makeFindCommand();
		
		ArrayList<ParseObject> answer = new ArrayList<ParseObject>();
		if (currentCommand == null) {
			return answer;
		}

		// Cache the results, unless we are ignoring the cache
		boolean caching = (cachePolicy != CachePolicy.IGNORE_CACHE);
		JSONObject result = (JSONObject) currentCommand.perform(caching);
		try {
			return convertFindResponse(result);
		} catch (JSONException e) {
			throw new RuntimeException(e.getMessage());
		}
	}
	
	/**
	 * Change the caching policy of this query.
	 */
	public void setCachePolicy(CachePolicy newCachePolicy) {
		checkIfRunning();

		cachePolicy = newCachePolicy;
	}
	
	/**
	 * Accessor for the caching policy.
	 */
	public CachePolicy getCachePolicy() {
		return cachePolicy;
	}
	
	/**
	 * Retrieves the results of the last time {@link find} was called on a query
	 * identical to this one.
	 * @return A list of ParseObjects corresponding to this query. Returns null if there is
	 * no cache for this query.
	 * @throws ParseException
	 */
	private ArrayList<ParseObject> findFromCache() throws ParseException {
		String cacheKey = makeFindCommand().getCacheKey();
		Object cached = Parse.jsonFromKeyValueCache(cacheKey);
		if (cached == null) {
			throw new ParseException(ParseException.CACHE_MISS, "results not cached");
		}
		if (!(cached instanceof JSONObject)) {
			throw new ParseException(ParseException.CACHE_MISS,
					"the cache contains the wrong datatype");
		}
		JSONObject object = (JSONObject) cached;
		try {
			return convertFindResponse(object);
		} catch (JSONException e) {
			throw new ParseException(ParseException.CACHE_MISS,
					"the cache contains corrupted json");
		}
	}

	/**
	 * Retrieves a list of ParseObjects that satisfy this query from the server
	 * in a background thread.
	 * This is preferable to using find(), unless your code is already running in
	 * a background thread.
	 * @param callback callback.done(objectList, e) is called when the find completes.
	 */
	public void findInBackground(FindCallback callback) {
		checkIfRunning(true);
		
		final CachePolicy callbackCachePolicy;

		if (cachePolicy == CachePolicy.CACHE_THEN_NETWORK) {
			// Do the cache callback on this thread immediately
			List<ParseObject> cacheResults;
			try {
				cacheResults = findWithCachePolicy(CachePolicy.CACHE_ONLY);
				callback.done(cacheResults, null);
			} catch (ParseException e) {
				callback.done(null, e);
			}
			
			callbackCachePolicy = CachePolicy.NETWORK_ONLY;
		} else {
			callbackCachePolicy = cachePolicy;
		}
		
		
		BackgroundTask<List<ParseObject>> findTask = new BackgroundTask<List<ParseObject>>(callback) {
			public List<ParseObject> run() throws ParseException {
				List<ParseObject> result = ParseQuery.this.findWithCachePolicy(callbackCachePolicy);
				currentCommand = null;
				return result;
			}
		};
		currentTask = findTask;
		BackgroundTask.executeTask(findTask);
	}
	
	/**
	 * Counts the number of objects that match this query.
	 * This does not use caching.
	 * @throws ParseException Throws an exception when the network connection fails or
	 * when the query is invalid.
	 */
	public int count() throws ParseException {
		return count(true);
	}
	
	protected int count(boolean needsLock) throws ParseException {
		if (needsLock) { checkIfRunning(true); }
		
		JSONObject result;
		
		try {
			currentCommand = makeFindCommand();
			// No need to return results
			currentCommand.put("limit", 0);
			currentCommand.put("count", 1);
			result = (JSONObject) currentCommand.perform();
		} finally {
			finishedRunning();
		}
		
		return result.optInt("count");
	}
	
	/**
	 * Counts the number of objects that match this query in a background thread.
	 * This does not use caching.
	 * @param callback callback.done(count, e) will be called when the count completes.
	 */
	public void countInBackground(CountCallback callback) {
		checkIfRunning(true);
		
		BackgroundTask<Integer> countTask = new BackgroundTask<Integer>(callback) {
			public Integer run() throws ParseException {
				Integer result = ParseQuery.this.count(false);
				currentCommand = null;
				return result;
			}
		};
		currentTask = countTask;
		BackgroundTask.executeTask(countTask);
	}
	
	/**
	 * Constructs a ParseObject whose id is already known by fetching data from the server.
	 * This does not use caching.
	 * @param theObjectId Object id of the ParseObject to fetch.
	 * @throws ParseException Throws an exception when there is no such object or when the
	 * network connection fails.
	 */
	public ParseObject get(String theObjectId) throws ParseException {
		return get(theObjectId, true);
	}
	
	protected ParseObject get(String theObjectId, boolean needsLock) throws ParseException {
		if (needsLock) { checkIfRunning(true); }
		
		ParseObject object;
		
		try {
			object = ParseObject.build(className);
			currentCommand = new ParseCommand("get");
			currentCommand.enableRetrying();
			currentCommand.put("classname", className);
			JSONObject data = new JSONObject();
			try {
				data.put("objectId", theObjectId);
			} catch (JSONException e) {
				throw new RuntimeException(e.getMessage());
			}
			currentCommand.put("data", data);
			JSONObject result = (JSONObject) currentCommand.perform();
			object.mergeObject(result, false);
		} finally {
			finishedRunning();
		}
		
		return object;
	}
	
	/**
	 * Constructs a ParseObject whose id is already known by fetching data from the
	 * server in a background thread.
	 * This does not use caching.
	 * This is preferable to using the ParseObject(className, objectId) constructor,
	 * unless your code is already running in a background thread.
	 * @param objectId Object id of the ParseObject to fetch.
	 * @param callback callback.done(object, e) will be called when the fetch completes.
	 */
	public void getInBackground(String objectId, GetCallback callback) {	
		checkIfRunning(true);
		
		final String finalObjectId = objectId;
		BackgroundTask<ParseObject> getTask = new BackgroundTask<ParseObject>(callback) {
			public ParseObject run() throws ParseException {
				ParseObject result = ParseQuery.this.get(finalObjectId, false);
				currentCommand = null;
				return result;
			}
		};
		currentTask = getTask;
		BackgroundTask.executeTask(getTask);
	}

	
	/**
	 * Add a constraint to the query that forces a particular key's value to be equal to
	 * the provided value.
	 * @param key The key to check.
	 * @param value The value that the ParseObject must contain.
	 * @deprecated Use {@link #whereEqualTo} instead.
	 */
	@Deprecated
	public void putWhere(String key, Object value) {
		checkIfRunning();

		whereEqualTo(key, value);
	}
	
	/**
	 * Add a constraint to the query that requires a particular key's value to be equal
	 * to the provided value.
	 * @param key The key to check.
	 * @param value The value that the ParseObject must contain.
	 * @return Returns the query, so you can chain this call.
	 */
	public ParseQuery whereEqualTo(String key, Object value) {
		checkIfRunning();
		
		if (value instanceof Date) {
			value = Parse.dateToObject((Date) value);
		}
		
		where.put(key, value);
		return this;
	}

	// Helper for condition queries.
	private void addCondition(String key, String condition, Object value) {
		checkIfRunning();

		JSONObject whereValue = null;

		value = Parse.maybeReferenceAndEncode(value);
		
		// Check if we already have some of a condition
		if (where.containsKey(key)) {
			Object existingValue = where.get(key);
			if (existingValue instanceof JSONObject) {
				whereValue = (JSONObject) existingValue;
			}
		}
		if (whereValue == null) {
			whereValue = new JSONObject();
		}
	
		try {
			whereValue.put(condition, value);
		} catch (JSONException e) {
			throw new RuntimeException(e.getMessage());
		}
		
		where.put(key, whereValue);
	}

	/**
	 * Add a constraint to the query that requires a particular key's value to be
	 * less than the provided value.
	 * @param key The key to check.
	 * @param value The value that provides an upper bound.
	 * @return Returns the query, so you can chain this call.
	 */
	public ParseQuery whereLessThan(String key, Object value) {
		addCondition(key, "$lt", value);
		return this;
	}
	
	/**
	 * Add a constraint to the query that requires a particular key's value to be
	 * not equal to the provided value.
	 * @param key The key to check.
	 * @param value The value that must not be equalled.
	 * @return Returns the query, so you can chain this call.
	 */
	public ParseQuery whereNotEqualTo(String key, Object value) {
		addCondition(key, "$ne", value);
		return this;
	}
	
	/**
	 * Add a constraint to the query that requires a particular key's value to be
	 * greater than the provided value.
	 * @param key The key to check.
	 * @param value The value that provides an lower bound.
	 * @return Returns the query, so you can chain this call.
	 */
	public ParseQuery whereGreaterThan(String key, Object value) {
		addCondition(key, "$gt", value);
		return this;
	}
	
	/**
	 * Add a constraint to the query that requires a particular key's value to be
	 * less than or equal to the provided value.
	 * @param key The key to check.
	 * @param value The value that provides an upper bound.
	 * @return Returns the query, so you can chain this call.
	 */
	public ParseQuery whereLessThanOrEqualTo(String key, Object value) {
		addCondition(key, "$lte", value);
		return this;
	}

	/**
	 * Add a constraint to the query that requires a particular key's value to be
	 * greater than or equal to the provided value.
	 * @param key The key to check.
	 * @param value The value that provides an lower bound.
	 * @return Returns the query, so you can chain this call.
	 */	
	public ParseQuery whereGreaterThanOrEqualTo(String key, Object value) {
		addCondition(key, "$gte", value);
		return this;
	}

	/**
	 * Add a constraint to the query that requires a particular key's value to be
	 * contained in the provided list of values.
	 * @param key The key to check.
	 * @param values The values that will match.
	 * @return Returns the query, so you can chain this call.
	 */
	public ParseQuery whereContainedIn(String key, Collection<? extends Object> values) {
		JSONArray array = new JSONArray();
		for (Object val : values) {
			array.put(Parse.maybeReferenceAndEncode(val));
		}
		addCondition(key, "$in", array);
		return this;
	}
	
	/**
	 * Add a proximity based constraint for finding objects with key point values near the
	 * point given.
	 * @param key The key that the ParseGeoPoint is stored in.
	 * @param point The reference ParseGeoPoint that is used.
	 * @return Returns the query, so you can chain this call.
	 */
	public ParseQuery whereNear(String key, ParseGeoPoint point) {
		addCondition(key, "$nearSphere", point);
		return this;
	}

	/**
	 * Add a proximity based constraint for finding objects with key point values near the
	 * point given and within the maximum distance given.
	 * Radius of earth used is 3958.8 miles.
	 * @param key The key that the ParseGeoPoint is stored in.
	 * @param point The reference ParseGeoPoint that is used.
	 * @param maxDistance Maximum distance (in miles) of results to return.
	 * @return Returns the query, so you can chain this call.
	 */
	public ParseQuery whereWithinMiles(String key, ParseGeoPoint point, double maxDistance) {
		whereWithinRadians(key, point, maxDistance / ParseGeoPoint.EARTH_MEAN_RADIUS_MILE);
		return this;
	}
	
	/**
	 * Add a proximity based constraint for finding objects with key point values near the
	 * point given and within the maximum distance given.
	 * Radius of earth used is 6371.0 kilometers.
	 * @param key The key that the ParseGeoPoint is stored in.
	 * @param point The reference ParseGeoPoint that is used.
	 * @param maxDistance Maximum distance (in kilometers) of results to return.
	 * @return Returns the query, so you can chain this call.
	 */
	public ParseQuery whereWithinKilometers(String key, ParseGeoPoint point, double maxDistance) {
		whereWithinRadians(key, point, maxDistance / ParseGeoPoint.EARTH_MEAN_RADIUS_KM);
		return this;
	}

	/**
	 * Add a proximity based constraint for finding objects with key point values near the
	 * point given and within the maximum distance given.
	 * @param key The key that the ParseGeoPoint is stored in.
	 * @param point The reference ParseGeoPoint that is used.
	 * @param maxDistance Maximum distance (in radians) of results to return.
	 * @return Returns the query, so you can chain this call.
	 */
	public ParseQuery whereWithinRadians(String key, ParseGeoPoint point, double maxDistance) {
		addCondition(key, "$nearSphere", point);
		addCondition(key, "$maxDistance", maxDistance);
		return this;
	}
		
	/**
	 * Add a regular expression constraint for finding string values that match
	 * the provided regular expression.
	 * This may be slow for large datasets.
	 * @param key The key that the string to match is stored in.
	 * @param regex The regular expression pattern to match.
	 * @return Returns the query, so you can chain this call.
	 */
	public ParseQuery whereMatches(String key, String regex) {
		addCondition(key, "$regex", regex);
		return this;
	}
	
	/**
	 * Add a constraint for finding string values that contain a provided string.
	 * This will be slow for large datasets. 
	 * @param key The key that the string to match is stored in.
	 * @param substring The substring that the value must contain.
	 * @return Returns the query, so you can chain this call.
	 */
	public ParseQuery whereContains(String key, String substring) {
		String regex = Pattern.quote(substring);
		whereMatches(key, regex);
		return this;
	}
	
	/**
	 * Add a constraint for finding string values that start with a provided string.
	 * This query will use the backend index, so it will be fast even for large datasets.
	 * @param key The key that the string to match is stored in.
	 * @param prefix The substring that the value must start with.
	 * @return Returns the query, so you can chain this call.
	 */
	public ParseQuery whereStartsWith(String key, String prefix) {
		String regex = "^" + Pattern.quote(prefix);
		whereMatches(key, regex);
		return this;
	}
	
	/**
	 * Add a constraint for finding string values that end with a provided string.
	 * This will be slow for large datasets.
	 * @param key The key that the string to match is stored in.
	 * @param suffix The substring that the value must end with.
	 * @return Returns the query, so you can chain this call.
	 */
	public ParseQuery whereEndsWith(String key, String suffix) {
		String regex = Pattern.quote(suffix) + "$";
		whereMatches(key, regex);
		return this;
	}
	
	/**
	 * Include nested ParseObjects for the provided key.
	 */
	public void include(String key) {
		checkIfRunning();
		
		include.add(key);
	}
	
	/**
	 * Controls the order in which ParseObjects are returned.
	 * The order may be the name of any field in the object, or the special strings
	 * <code>"_updated_at"</code> or <code>"_created_at"</code>
	 * to sort by update or creation times.
	 * <p>
	 * The order may also be prefixed with a "-" to reverse the order.
	 * @param newOrder
	 * @deprecated Use {@link #orderByAscending(String)} or {@link #orderByDescending(String)} instead.
	 */
	public void setOrder(String newOrder) {
		checkIfRunning();

		order = newOrder;
	}
	
	/**
	 * Sorts the results in ascending order by the given key.
	 * 
	 * @param key The key to order by.
	 * @return Returns the query, so you can chain this call. 
	 */
	public ParseQuery orderByAscending(String key) {
		checkIfRunning();
		
		order = key;
		return this;		
	}
	
	/**
	 * Sorts the results in descending order by the given key.
	 * 
	 * @param key The key to order by.
	 * @return Returns the query, so you can chain this call.
	 */
	public ParseQuery orderByDescending(String key) {
		checkIfRunning();
		
		order = "-" + key;
		return this;
	}
	
	/**
	 * Controls the maximum number of results that are returned.
	 * Setting a negative limit denotes retrieval without a limit.
	 * @param newLimit
	 */
	public void setLimit(int newLimit) {
		checkIfRunning();

		limit = newLimit;
	}
	
	/**
	 * Accessor for the limit.
	 */
	public int getLimit() {
		return limit;
	}
	
	/**
	 * Controls the number of results to skip before returning any results.
	 * This is useful for pagination.
	 * Default is to skip zero results.
	 * @param newSkip
	 */
	public void setSkip(int newSkip) {
		checkIfRunning();

		skip = newSkip;
	}
	
	/**
	 * Accessor for the skip value.
	 */
	public int getSkip() {
		return skip;
	}
	
	/**
	 * Accessor for the class name.
	 */
	public String getClassName() {
		return className;
	}
}
