package com.parse;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.net.ServerSocket;
import java.util.concurrent.CountDownLatch;
import java.util.concurrent.Semaphore;
import java.util.concurrent.TimeUnit;

import org.json.JSONException;

import android.content.Context;
import android.os.Looper;
import android.util.Log;

/**
 * A helper object for testing that asynchronous calls actually happen.
 * When you call assertFinishes() in one thread, it will wait for another
 * thread to call finish().
 * It's easiest to use this in conjunction with runAsyncTest.
 */
class Synchronizer {
	boolean done;
	CountDownLatch latch;
	Synchronizer() {
		done = false;
		// Implicit use case
		latch = new CountDownLatch(1);
	}
	
	void start(int count) {
		latch = new CountDownLatch(count);
	}
	
	void finish() {
		if (Thread.currentThread() != Looper.getMainLooper().getThread()) {
			throw new RuntimeException("finish() should be called from the UI thread");
		}
		
		if (done) {
			throw new RuntimeException("finish() was called too much");
		}

		latch.countDown();
		
		if (latch.getCount() == 0) {
			done = true;			
		}
	}
	
	void assertFinishes() {		
		if (Thread.currentThread() == Looper.getMainLooper().getThread()) {
			throw new RuntimeException(
					"assertFinishes() should *not* be called from the UI thread");
		}
		try {
			// Allow outstanding callbacks a chance to resolve.  This delay was
			// chosen after some experimentation but does not guarantee that all
			// possible callbacks have been resolved.  If we don't care about calling
			// finish excessively, this can be removed.  yield() is not sufficient.
			Thread.sleep(500);
			latch.await(10, TimeUnit.SECONDS);
		} catch (InterruptedException e) {
			throw new RuntimeException(e.getMessage());
		}

		if (!done) {
			throw new RuntimeException("finish() was never called");
		}
	}
}

public class ParseTestUtils {
	private static final String TAG = "com.parse.ParseTestUtils";
	
	/**
	 * Use a different backend for the Parse service.
	 * In general, you do not need to call this.
	 * @param theServer Root URL of the server
	 */
	public static void useServer(String theServer) {
		ParseObject.server = theServer;
	}
	
	/**
	 * Use an invalid server name.
	 * This ensures that we don't access it.
	 */
	public static void useInvalidServer() {
		useServer("http://invalid.server:3000");
	}
	
	/**
	 * Use a port with nothing listening.
	 * This ensures we get a network failure when we access it.
	 */
	public static void useBadServerPort() {
		useServer("http://10.0.2.2:6666");
	}
	
	/**
	 * Only works on the test database.
	 */
	public static void clearApp() {
		ParseCommand command = new ParseCommand("clear_app");
		try {
			command.perform();
		} catch (ParseException e) {
			throw new RuntimeException(e.getMessage());
		}
	}
	
	public static void beginFakeSleep() {
		PushService.sleepSemaphore = new Semaphore(0);
	}
	
	public static void endFakeSleep() {
		PushService.sleepSemaphore.release(1000000);
	}
	
	public static void allowSleep(int millis) {
		PushService.sleepSemaphore.release(millis);
	}
	
	// Asserts that all the permitted sleep will be slept.
	public static void assertSlept() {
		while (true) {
			if (PushService.sleepSemaphore.availablePermits() == 0) {
				return;
			}
			try {
				Thread.sleep(50);
			} catch (InterruptedException e) {
				throw new RuntimeException(e.getMessage());
			}
		}
	}
	
	// The server port changes so that successive test runs don't interfere
	// with each other.
	private static int serverPort = 9000;
	
	/**
	 * Points to a push server on the development machine, assuming this is the emulator.
	 */
	public static void useDevPushServer() {
		PushService.useServer("10.0.2.2");
	}
	
	public static int consecutiveFailures() {
		return PushService.consecutiveFailures;
	}
	
	public static void saveObjectToDisk(ParseObject object, Context context, String filename) {
		object.saveToDisk(context, filename);
	}
	
	public static ParseObject getObjectFromDisk(Context context, String filename) {
		return ParseObject.getFromDisk(context, filename);
	}
	
	public static ParseUser getUserObjectFromDisk(Context context, String filename) {
		return (ParseUser)ParseUser.getFromDisk(context, filename);
	}
	
	public static void saveStringToDisk(String string, Context context, String filename) {
		File file = new File(getParseDir(context), filename);
		try {
			FileOutputStream out = new FileOutputStream(file);
			out.write(string.getBytes("UTF-8"));
			out.close();
		} catch (UnsupportedEncodingException e) {
			return;
		} catch (IOException e) {
			return;
		}
	}
	
	static File getParseDir(Context context) {
		return context.getDir("Parse", 0);
	}
	
	
	private static Synchronizer synchronizer;

	public static void initSynchronizer() {
		synchronizer = new Synchronizer();
	}
	
	// Sets up the latch in the synchronizer.
	public static void start(int count) {
		synchronizer.start(count);
	}
		
	// Call assertFinishes() from the testing thread.
	public static void assertFinishes() {
		synchronizer.assertFinishes();
	}
	
	// Call finish() from the UI thread.
	public static void finish() {
		synchronizer.finish();
	}
	
	public static void clearInstallationId() {
		ParseObject.installationId = null;
	}
	
	public static void recursiveDelete(File file) {
		if (!file.exists()) {
			return;
		}
		if (file.isDirectory()) {
			// Delete kids
			File[] children = file.listFiles();
			for (File child : children) {
				recursiveDelete(child);
			}
		}
		file.delete();
	}
	
	public static void clearFiles() {
		Log.d(TAG, "clearFiles()");
		recursiveDelete(Parse.getParseDir());
		recursiveDelete(Parse.getKeyValueCacheDir());
	}
	
	public static void clearPushServiceCallbacks() {
		PushService.transientCallbacks = null;
		PushService.persistentCallbacks = null;
	}
	
	public static void onPush(Context context, String channel, PushCallback callback) {
		PushService.addTransientPushCallback(context, channel, callback);
	}
	
	public static int totalNotifications() {
		return StandardPushCallback.totalNotifications;
	}
	
	public static String getInstallationId(Context context) {
		return ParseObject.getInstallationId(context);
	}
	
	public static String getLastTime() {
		return PushService.lastTime;
	}
	
	// Various helpers for the PushTest teardown that need to run here so that
	// they can access namespace-private things
	public static void tearDownPushTest() {
		clearFiles();
		clearInstallationId();
		clearPushServiceCallbacks();
		PushService.socket = null;
	}
	
	// Various helpers for the PushTest setup.
	public static void setUpPushTest() {
		// It's nice if we disable the actual showing of a notification in the
		// tray, because during unit tests that is slow and we can't see it
		// anyway.
		StandardPushCallback.disableNotifications = true;
		
		StandardPushCallback.totalNotifications = 0;
		PushService.sleepSemaphore = null;
		PushService.lastTime = null;
		PushService.useServer("localhost");
		useServer("The ParseObject server should not be used");		
		ParseTestUtils.initSynchronizer();		
	}
	
	public static void startServiceIfRequired(Context context) {
		PushService.startServiceIfRequired(context);
	}
	
	public static String toDeterministicString(Object o) {
		try {
			return ParseCommand.toDeterministicString(o);
		} catch (JSONException e) {
			throw new RuntimeException(e.getMessage());
		}
	}
	
	/**
	 * Points to a push server on the emulator itself, and creates a socket
	 * for the other end of that push server.
	 */
	public static ServerSocket mockPushServer() {
		++serverPort;
		PushService.usePort(serverPort);
		Log.d(TAG, "running mockPushServer on port " + serverPort);
		try {
			return new ServerSocket(serverPort);
		} catch (IOException e) {
			throw new RuntimeException(e.getMessage());
		}
	}
	
	/**
	 *  How many files are in the key-value cache.
	 */
	public static int numKeyValueCacheFiles() {
		return Parse.getKeyValueCacheDir().listFiles().length;
	}
	
	/**
	 *  Set the maximum number of files for the key-value cache
	 */
	public static void setMaxKeyValueCacheFiles(int max) {
		Parse.maxKeyValueCacheFiles = max;
	}
	
	/**
	 * Set the maximum number of bytes for the key-value cache
	 */
	public static void setMaxKeyValueCacheBytes(int max) {
		Parse.maxKeyValueCacheBytes = max;
	}
}