package com.parse;

import java.io.File;

import org.json.JSONException;
import org.json.JSONObject;

public class ParseUser extends ParseObject {
	
	private String password;
	private String sessionToken;
	
	private static ParseUser currentUser;
	
	// Whether currentUser is known to match the serialized version on disk.
	// This is useful for saving a filesystem check if you try to load currentUser
	// frequently while there is none on disk.
	private static boolean currentUserMatchesDisk = false;
	
	private final static String CURRENT_USER_FILENAME = "currentUser";
	private final static String CLASS_NAME = ParseObject.USER_CLASS_NAME;
	
	// Whether the object is a currentUser. If so, it will always be persisted to disk on updates.
	private boolean isCurrentUser; 

	/**
	 * Constructs a new ParseUser with no data in it.
	 * A ParseUser constructed in this way will not have 
	 * an objectId and will not persist to the database 
	 * until {@link #signUp} is called.
	 */
	public ParseUser() {
		super(CLASS_NAME);
		isCurrentUser = false;
	}
	
	/**
	 * Whether the ParseUser has been authenticated on this device. This will be true
	 * if the ParseUser was obtained via a logIn or signUp method. Only an authenticated
	 * ParseUser can be saved (with altered attributes) and deleted.
	 */
	public boolean isAuthenticated() {
		return sessionToken != null;
	}
	
	public void remove(String key) {
		if (key == "username") {
			throw new IllegalArgumentException("Can't remove the username key.");
		}
		super.remove(key);
	}
	
	/**
	 * Need to send sessionToken if the user has one.
	 * So the structure of the ParseUser JSON representation could a "sessionToken" key.
	 */
	JSONObject toJSONObject(boolean includeDates, boolean allKeys) {
		JSONObject objectJSON = super.toJSONObject(includeDates, allKeys);
		
		if (sessionToken != null) {
			try {
				objectJSON.put("session_token", sessionToken);
			} catch (JSONException e) {
				throw new RuntimeException("could not encode value for key: sessionToken");
			}
		}
		
		return objectJSON;
	}

	// Additional functionality needed for merging values from server
	void mergeObject(JSONObject object, boolean justCreated) {
		super.mergeObject(object, justCreated);
		
		// Merge in the sessionToken, if it exists
		if (object.has("session_token")) {
			try {
				sessionToken = object.getString("session_token");
			} catch (JSONException e) {
				throw new RuntimeException(e.getMessage());
			}
		}
	}
	
	/**
	 * Sets the username. Usernames cannot be null or blank.
	 * @param username The username to set.
	 */
	public void setUsername(String username) {
		checkIfRunning();
		put("username", username);
	}

	/**
	 * Retrieves the username.
	 */
	public String getUsername() {
		return getString("username");
	}
	
	/**
	 * Sets the password.
	 * @param password The password to set.
	 */
	public void setPassword(String password) {
		checkIfRunning();
		this.password = password;
		dirty = true;
	}
	
	/**
	 * Sets the email address.
	 * @param email The email address to set.
	 */
	public void setEmail(String email) {
		checkIfRunning();
		put("email", email);
	}
	
	/**
	 * Retrieves the email address.
	 */
	public String getEmail() {
		return getString("email");
	}
	
	/**
	 * Retrieves the session token for a user, if they are logged in.
	 */
	public String getSessionToken() {
		return sessionToken;
	}
	
	void validateSave() {
		if (getObjectId() == null) {
			throw new IllegalArgumentException("Cannot save a ParseUser until it has been signed up. Call signUp first.");
		}
		
		if (!isAuthenticated() && isDirty()) {
			throw new IllegalArgumentException("Cannot save a ParseUser that is not authenticated.");
		}
	}
	
	public void save() throws ParseException {
		super.save();
		
		// If the user is the currently logged in user, we persist all data to disk
		if (isCurrentUser) {
			ParseUser.saveCurrentUser(this);
		}
	}
	
	void validateDelete() {
		if (!isAuthenticated() && isDirty()) {
			throw new IllegalArgumentException("Cannot delete a ParseUser that is not authenticated.");
		}
	}
	
	public void refresh() throws ParseException {
		super.refresh();
		
		if (isCurrentUser) {
			ParseUser.saveCurrentUser(this);
		}
	}
	
	ParseCommand constructSaveCommand() throws ParseException {
		ParseCommand command = super.constructSaveCommand();
		
		if (command == null) {
			return null;
		}
		
		if (password != null) {
			command.put("user_password", password);
		}
		
		if (sessionToken != null) {
			command.put("session_token", sessionToken);
		}
		
		return command;
	}
	
	ParseCommand constructDeleteCommand() throws ParseException {
		ParseCommand command = super.constructDeleteCommand();
		if (sessionToken != null) {
			command.put("session_token", sessionToken);
		}
		return command;
	}
	
	private ParseCommand constructSignUpCommand() throws ParseException {
		ParseCommand command = constructSaveCommand();
		command.setOp("user_signup");
		return command;
	}
	
	private static ParseCommand constructPasswordResetCommand(String email) {
		ParseCommand command = new ParseCommand("user_request_password_reset");
		command.put("email", email);
		return command;
	}

	
	/**
	 * Signs up a new user. You should call this instead of {@link #save} for new
	 * ParseUsers. This will create a new ParseUser on the server, and also persist
	 * the session on disk so that you can access the user using {@link #getCurrentUser}.
	 * 
	 * <p>A username and password must be set before calling signUp.</p>
	 * 
	 * <p>Typically, you should use {@link #signUpInBackground} instead of this,
	 * unless you are managing your own threading.</p>
	 *  
	 * @throws ParseException Throws an exception if the server is inaccessible, or if
	 * the username has already been taken.
	 */
	public void signUp() throws ParseException {
		signUp(true);
	}
	
	protected void signUp(boolean needsLock) throws ParseException {
		if (needsLock) { checkIfRunning(true); }
		
		try {
			if (getObjectId() != null) {
				throw new IllegalArgumentException("Cannot sign up a user that has already signed up.");
			}
			
			if (getUsername() == null || getUsername().length() == 0) {
				throw new IllegalArgumentException("Username cannot be missing or blank");
			}
	
			if (password == null) {
				throw new IllegalArgumentException("Password cannot be missing or blank");
			}
			
			ParseCommand command = constructSignUpCommand();
			if (command == null) {
				return;
			}
			
			JSONObject result = (JSONObject) command.perform();
			
			handleSaveResult(command.op, result);
			
			ParseUser.saveCurrentUser(this);
		} finally {
			finishedRunning();
		}
	}
	
	/**
	 * Signs up a new user. You should call this instead of {@link #save} for new
	 * ParseUsers. This will create a new ParseUser on the server, and also persist
	 * the session on disk so that you can access the user using {@link #getCurrentUser}.
	 * 
	 * <p>A username and password must be set before calling signUp.</p>
	 * 
	 * <p>This is preferable to using {@link #signUp}, unless your code is
	 * already running from a background thread.
	 *  
	 * @param callback callback.done(user, e) is called when the signUp completes.
	 * @throws ParseException Throws an exception if the server is inaccessible, or if
	 * the username has already been taken.
	 */
	public void signUpInBackground(SignUpCallback callback) {
		checkIfRunning(true);
		
		BackgroundTask<Void> signUpTask = new BackgroundTask<Void>(callback) {
			public Void run() throws ParseException {
				ParseUser.this.signUp(false);
				return null;
			}
		};
		BackgroundTask.executeTask(signUpTask);
	}
	
	private static ParseCommand constructLogInCommand(String username, String password) throws ParseException {
		ParseCommand command = new ParseCommand("user_login");
		command.put("username", username);
		command.put("user_password", password);
		
		return command;
	}
	
	/**
	 * Logs in a user with a username (or email) and password. On success, this saves
	 * the session to disk, so you can retrieve the currently logged in user
	 * using {@link #getCurrentUser} 
	 * 
	 * <p>Typically, you should use {@link #logInInBackground} instead of this,
	 * unless you are managing your own threading.</p>
	 *  
	 * @param username The username (or email) to log in with.
	 * @param password The password to log in with.
	 * @throws ParseException Throws an exception if the server is inaccessible. 
	 * @return The user if the login was successful. If the log in was unsuccessful, null will be returned.
	 */
	public static ParseUser logIn(String username, String password) throws ParseException {
		if (username == null) {
			throw new IllegalArgumentException("Must specify a username for the user to log in with");
		}
		
		if (password == null) {
			throw new IllegalArgumentException("Must specify a password for the user to log in with");
		}
		
		ParseCommand command = constructLogInCommand(username, password);
		
		Object result = command.perform();
		
		if (result == JSONObject.NULL) {
			return null;
		} else {
			ParseUser user = new ParseUser();
			user.handleSaveResult(command.op, (JSONObject) result);
			ParseUser.saveCurrentUser(user);
			return user;
		}
	}
	
	/**
	 * Logs in a user with a username (or email) and password. On success, this saves
	 * the session to disk, so you can retrieve the currently logged in user
	 * using {@link #getCurrentUser} 
	 * 
	 * <p>This is preferable to using {@link #logIn}, unless your code is
	 * already running from a background thread.</p>
	 *  
	 * @param username The username (or email) to log in with.
	 * @param password The password to log in with.
	 * @param callback callback.done(user, e) is called when the login completes.
	 * @throws ParseException Throws an exception if the server is inaccessible. 
	 */
	public static void logInInBackground(final String username, final String password, LogInCallback callback) {
		BackgroundTask<ParseUser> logInTask = new BackgroundTask<ParseUser>(callback) {
			public ParseUser run() throws ParseException {
				ParseUser user = ParseUser.logIn(username, password);
				return user;
			}
		};
		BackgroundTask.executeTask(logInTask);
	}
	
	/**
	 * This retrieves the currently logged in ParseUser with a valid session,
	 * either from memory or disk if necessary.
	 * @return The currently logged in ParseUser
	 */
	public static ParseUser getCurrentUser() {
		checkApplicationContext();
		
		if (currentUser != null) {
			return currentUser;
		}
		if (currentUserMatchesDisk) {
			return null;
		}
		
		// Load the current user from the disk, so after this call, current user
		// will match disk.
		currentUserMatchesDisk = true;
		ParseObject user = getFromDisk(Parse.applicationContext, CURRENT_USER_FILENAME);
		if (user == null) {
			return null;
		}
		currentUser = (ParseUser) user;
		currentUser.isCurrentUser = true;
		return currentUser;
	}
	
	// Persists a user as currentUser to disk, and into the singleton
	private static void saveCurrentUser(ParseUser user) {
		checkApplicationContext();
		user.isCurrentUser = true;
		user.saveToDisk(Parse.applicationContext, CURRENT_USER_FILENAME);
		currentUserMatchesDisk = true;
		currentUser = user;
	}
	
	/**
	 * Logs out the currently logged in user session. This will remove the
	 * session from disk, and future calls to {@link #getCurrentUser()} will
	 * return null.
	 */
	public static void logOut() {
		checkApplicationContext();
		currentUserMatchesDisk = true;
		currentUser = null;
		File file = new File(Parse.getParseDir(), CURRENT_USER_FILENAME);
		file.delete();
	}
	
	
	/**
	 * Requests a password reset email to be sent to the specified email address
	 * associated with the user account. This email allows the user to securely
	 * reset their password on the Parse site.
	 * 
	 * <p>Typically, you should use {@link #requestPasswordResetInBackground} instead of this,
	 * unless you are managing your own threading.</p>
	 * 
	 * @param email The email address associated with the user that forgot their password.
	 * @throws ParseException Throws an exception if the server is inaccessible, or if an account with that email doesn't exist.
	 */
	public static void requestPasswordReset(String email) throws ParseException {
		ParseCommand command = constructPasswordResetCommand(email);
		command.perform();
	}
	
	/**
	 * Requests a password reset email to be sent in a background thread to the specified email address
	 * associated with the user account. This email allows the user to securely
	 * reset their password on the Parse site.
	 * 
	 * <p>This is preferable to using requestPasswordReset(), unless your code is
	 * already running from a background thread.</p>
	 * 
	 * @param email The email address associated with the user that forgot their password.
	 * @param callback callback.done(e) is called when the request completes.
	 */
	public static void requestPasswordResetInBackground(final String email, RequestPasswordResetCallback callback) {
		BackgroundTask<Void> task = new BackgroundTask<Void>(callback) {
			public Void run() throws ParseException {
				ParseUser.requestPasswordReset(email);
				return null;
			}
		};
		BackgroundTask.executeTask(task);
	}
	
	/** 
	 * Checks to make sure that there is a valid applicationContext. In prior 
	 * versions, a context was not required to be pass into Parse.initialize.
	 */
	private static void checkApplicationContext() {
		if (Parse.applicationContext == null) {
			throw new RuntimeException(
					"You must call Parse.initialize(context, oauthKey, oauthSecret)"
					+ " before using the Parse library.");
		}
	}
}
