package com.parse;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.io.UnsupportedEncodingException;
import java.net.InetSocketAddress;
import java.net.Socket;
import java.net.UnknownHostException;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;
import java.util.Timer;
import java.util.TimerTask;
import java.util.concurrent.Semaphore;
import java.util.regex.Pattern;

import org.json.JSONArray;
import org.json.JSONException;
import org.json.JSONObject;
import org.json.JSONTokener;

import android.app.Activity;
import android.app.Service;
import android.content.ComponentName;
import android.content.Context;
import android.content.Intent;
import android.content.pm.ApplicationInfo;
import android.content.pm.PackageManager;
import android.content.pm.PackageManager.NameNotFoundException;
import android.os.IBinder;
import android.util.Log;

import com.parse.os.ParseAsyncTask;

/**
 * A service to listen for push notifications.
 * This operates in the same process as the parent application.
 * 
 * To use this class, the PushService must be registered. Add this XML right before the
 * <code>&lt;/application&gt;</code> tag in your AndroidManifest.xml:
 * <pre>
 * &lt;service android:name="com.parse.PushService" /&gt;
 * &lt;receiver android:name="com.parse.ParseBroadcastReceiver"&gt;
 *   &lt;intent-filter&gt;
 *     &lt;action android:name="android.intent.action.BOOT_COMPLETED" /&gt;
 *     &lt;action android:name="android.intent.action.USER_PRESENT" /&gt;
 *   &lt;/intent-filter&gt;
 * &lt;/receiver&gt;
 * </pre>
 * 
 * Next, you must ensure your app has the permissions needed to show a notification.
 * Make sure that these permissions are present in your AndroidManifest.xml, typically
 * immediately before the <code>&lt;/manifest&gt;</code> tag:
 * 
 * <pre>
 * &lt;uses-permission android:name="android.permission.INTERNET" /&gt;
 * &lt;uses-permission android:name="android.permission.RECEIVE_BOOT_COMPLETED" /&gt;
 * &lt;uses-permission android:name="android.permission.VIBRATE" /&gt;
 * </pre>
 * 
 * Once push notifications are configured in the manifest, you can subscribe to a push
 * channel by calling
 * 
 * <pre>
 * PushService.subscribe(context, "the_channel_name", YourActivity.class);
 * </pre>
 * 
 * When the client receives a push message, a notification will appear in the system tray.
 * When the user taps the notification, they will enter the application through a new 
 * instance of YourActivity.
 */
public final class PushService extends Service {
	private static final String TAG = "com.parse.PushService";
	private static String pushServer = "push.parse.com";
	private static int defaultPushPort = 8253;
	private int pushPort;
	
	// Callbacks that will disappear when the service is killed.
	// Since a non-service thread could add a callback here, access to
	// transientCallbacks should only happen within synchronized blocks on
	// PushService.class.
	static HashMap<String, PushCallback> transientCallbacks;
	
	// The persistent callback data that has been serialized to disk
	// Since a non-service thread could add a callback here, access to
	// persistentCallbacks should only happen within synchronized blocks on
	// PushService.class.
	static JSONObject persistentCallbacks;
    
	// The socket we use for communicating with the push server.
	// This socket is used from multiple threads at once, sending on
	// one thread while reading on another thread and perhaps closing the
	// socket from yet another thread. That is theoretically
	// dangerous. In practice it does not seem to cause problems. However
	// in the long run it would be a good idea to clean that up.
	static Socket socket;
	
	// The task used for background processing.
	private ParseAsyncTask<Void, Void, String> task;
	
	// The amount of time to delay after the next failure
	private static int delaySeconds;
	
	// Whether this Service has been destroyed.
	// This is better to use than task.isCancelled() because there
	// are multiple tasks.
	private boolean destroyed = false;
	
	// The last timestamp we have received from the server, or null if there has been none.
	static String lastTime = null;
	
	// A timer for handling keepalives.
	private Timer keepAliveTimer = null;
	
	// Number of consecutive failures
	static int consecutiveFailures;
	
	// Used to fake sleeping, with each permit representing a millisecond.
	// Managed by ParseTestUtils.
	static Semaphore sleepSemaphore;
	
	/**
	 * Client code should not construct a PushService directly. 
	 */
	public PushService() {
		super();
	}
	
	// Start the PushService. Logs nicely if it can't.
	private static void startService(Context context) {
		Log.d(TAG, "ensuring push service is started");
		// This is a no-op if the service is already running.
		ComponentName name = context.startService(new Intent(context, PushService.class));
		if (name == null) {
			Log.e(TAG, "Could not start the push service. Make sure that the XML tag " +
					"<service android:name=\"com.parse.PushService\" /> is in your " +
					"AndroidManifest.xml as a child of the <application> element.");
		}		
	}
	
	// Start the PushService if there are any subscriptions.
	static synchronized void startServiceIfRequired(Context context) {
		loadPersistentCallbacks(context);
		if (persistentCallbacks.length() > 0 ||
				(transientCallbacks != null && transientCallbacks.size() > 0)) {
			startService(context);
		}
	}
	
	// Call this whenever the set of callbacks has changed.
	// This returns immediately and uses a thread to do the work, so that it
	// does not block.
	private static void callbacksChanged(Context context) {
		final Context finalContext = context;
		Runnable job = new Runnable() {
			public void run() {
				sendSubscriptionInformation(finalContext);
				startService(finalContext);
			}
		};
		Thread thread = new Thread(job);
		thread.start();
	}
	
	// Loads the persistent callback data from disk if necessary.
	private static synchronized void loadPersistentCallbacks(Context context) {
		if (persistentCallbacks == null) {
			persistentCallbacks = ParseObject.getDiskObject(context, "persistentCallbacks");
			if (persistentCallbacks == null) {
				persistentCallbacks = new JSONObject();
			}
		}
	}
	
	private static synchronized void savePersistentCallbacks(Context context) {
		if (persistentCallbacks != null) {
			ParseObject.saveDiskObject(context, "persistentCallbacks", persistentCallbacks);
		}
	}
	
	/**
	 * Listens for a push message on the specified channel until this service is destroyed.
	 * When such a message is received, the PushCallback gets called.
	 * 
	 * This specifically does not persist when the service gets shut down and restarted.
	 * For persistent callbacks, use addPersistentPushCallback.
	 * If there is both a persistent and a transient callback for a particular channel,
	 * the persistent callback takes precedence. That is a bit odd but transient
	 * callbacks are not yet exposed to end users so it's not so bad. 
	 * 
	 * @param context This is used to access local storage and activate the service, so
	 * it must currently be a viable context.
	 * @param channel Specifies which messages will be handled by this callback.
	 * @param callback The callback that gets run when the push happens.
	 */
	static void addTransientPushCallback(
			Context context, String channel, PushCallback callback) {
		synchronized (PushService.class) {
			if (transientCallbacks == null) {
				transientCallbacks = new HashMap<String, PushCallback>();
			}
			transientCallbacks.put(channel, callback);
		}
		callbacksChanged(context);
	}
	
	/**
	 * Like addTransientPushCallback, but serializes the callback to disk so that
	 * it can run even after the service restarts.
	 */
	static void addPersistentPushCallback(
			Context context, String channel, JSONObject localData,
			Class<? extends PushCallback> cls) {
		synchronized (PushService.class) {
			loadPersistentCallbacks(context);
			String callbackName = cls.getCanonicalName();
			JSONObject callbackData = new JSONObject();
			try {
				callbackData.put("data", localData);
				callbackData.put("name", callbackName);
				// Make a copy so that the caller can't futz with our internals
				JSONObject callbackDataCopy = new JSONObject(callbackData.toString());
				persistentCallbacks.put(channel, callbackDataCopy);
			} catch (JSONException e) {
				Log.d(TAG, "json exception subscribing to " + channel);
				return;
			}
			savePersistentCallbacks(context);
		}

		callbacksChanged(context);
	}
	
	/**
	 * Helper function to subscribe to push notifications with the default application icon.
	 * @param context This is used to access local storage to cache the subscription,
	 * so it must currently be a viable context.
	 * @param channel A string identifier that determines which messages will cause a push
	 * notification to be sent to this client.
	 * @param cls This should be a subclass of Activity. An instance of this Activity
	 * is started when the user responds to this push notification. If you are not sure what
	 * to use here, just use your application's main Activity subclass.
	 */
	public static void subscribe(
			Context context, String channel, Class<? extends Activity> cls) {
		// Load the application info
		String packageName = context.getPackageName();
		PackageManager pm = context.getPackageManager();
		ApplicationInfo info;
		try {
			info = pm.getApplicationInfo(packageName, 0);
		} catch (NameNotFoundException e) {
			Log.d(TAG, "missing package " + packageName);
			return;
		}
		subscribe(context, channel, cls, info.icon);
	}
	
	/**
	 * Call this function when the user should be subscribed to a new push channel.
	 * When push notifications are sent out on this channel subsequently, this device
	 * will display a "toast" notification in the system tray.
	 * This function returns immediately, even when there is no internet access. In that case the
	 * subscription is cached and when internet access is next available the client will
	 * inform the server of its subscription.
	 * This starts a PushService running in the background that will not permanently
	 * terminate as long as the user is subscribed to some channel, unless the application
	 * is uninstalled.
	 * 
	 * @param context This is used to access local storage to cache the subscription,
	 * so it must currently be a viable context.
	 * @param channel A string identifier that determines which messages will cause a push
	 * notification to be sent to this client.
	 * @param cls This should be a subclass of Activity. An instance of this Activity
	 * is started when the user responds to this push notification. If you are not sure what
	 * to use here, just use your application's main Activity subclass.
	 * @param icon The icon to show for the notification.
	 */
	public static void subscribe(
			Context context, String channel, Class<? extends Activity> cls, int icon) {
		// Ensure the channel is valid
		Pattern channelPattern = Pattern.compile("^[A-Za-z0-9_-]*$");
		if (!channelPattern.matcher(channel).matches()) {
			throw new RuntimeException("invalid channel name: " + channel);
		}
		
		// Cache the oauth key if we haven't already, so that we can restart later
		getApplicationId(context);
		
		// Load the application info
		String packageName = context.getPackageName();
		PackageManager pm = context.getPackageManager();
		ApplicationInfo info;
		try {
			info = pm.getApplicationInfo(packageName, 0);
		} catch (NameNotFoundException e) {
			Log.d(TAG, "missing package " + packageName);
			return;
		}
		
		// Get the app name
		CharSequence appChars = info.loadLabel(pm);
		final String appName;
		if (appChars != null) {
			appName = appChars.toString();
		} else {
			appName = null;
		}
		
		// We need the parts of the ComponentName so that we can fire off an
		// appropriate intent, later.
		ComponentName componentName = new ComponentName(context, cls);
		String activityClass = componentName.getClassName();
		String activityPackage = componentName.getPackageName();

		// Wrap up the data the PushCallback will need
		JSONObject localData = new JSONObject();
		try {
			localData.put("icon", icon);
			localData.put("appName", appName);		
			localData.put("activityClass", activityClass);
			localData.put("activityPackage", activityPackage);
		} catch (JSONException e) {
			// This should never happen
			throw new RuntimeException(e.getMessage());
		}
		
		addPersistentPushCallback(context, channel, localData, StandardPushCallback.class);
	}
	
	/**
	 * Cancels a previous call to subscribe.
	 * If the user is not subscribed to this channel, this is a no-op.
	 * This call does not require internet access. It returns without blocking.
	 * @param context A currently viable Context.
	 * @param channel The string defining the channel to unsubscribe from.
	 */
	public static synchronized void unsubscribe(Context context, String channel) {
		if (transientCallbacks != null) {
			transientCallbacks.remove(channel);
		}
		loadPersistentCallbacks(context);
		if (persistentCallbacks != null) {
			if (persistentCallbacks.has(channel)) {
				persistentCallbacks.remove(channel);
				savePersistentCallbacks(context);
			}
		}
		callbacksChanged(context);
	}
	
	/**
	 * An accessor to determine which channels this application is currently subscribed to.
	 * The local cache of subscriptions is always the authoritative one, so this function
	 * returns without blocking on network IO.
	 * @param context A currently viable Context.
	 * @return A set containing all of the channels this application is currently subscribed to.
	 */
	public static synchronized Set<String> getSubscriptions(Context context) {
		loadPersistentCallbacks(context);
		Set<String> answer = new HashSet<String>();
		@SuppressWarnings("unchecked")		
		Iterator<String> iter = persistentCallbacks.keys();
		while (iter.hasNext()) {
			answer.add(iter.next());
		}
		return answer;
	}

	// Called when we just received a push message.
	private void handlePushData(String channel, JSONObject pushData) {
		// Check persistent callbacks.
		// This block needs to be synchronized because it
		// interoperates with the callback maps.
		PushCallback callback = null;
		synchronized (PushService.class) {
			loadPersistentCallbacks(this);
			JSONObject callbackData = persistentCallbacks.optJSONObject(channel); 
			if (callbackData != null) {
				Class<?> c;
				String callbackName = callbackData.optString("name");
				try {
					c = Class.forName(callbackName);
				} catch (ClassNotFoundException e) {
					Log.d(TAG, "class not found: " + callbackName);
					return;
				}

				try {
					callback = (PushCallback) c.newInstance();
				} catch (IllegalAccessException e) {
					Log.d(TAG, "illegal access to " + callbackName);
					return;
				} catch (InstantiationException e) {
					Log.d(TAG, "could not instantiate " + callbackName);
					return;
				}
				// Use a copy of the local data to isolate our internal state
				JSONObject localData = callbackData.optJSONObject("data");
				JSONObject localDataCopy = null;
				if (localData != null) {
					try {
						localDataCopy = new JSONObject(localData.toString());
					} catch (JSONException e) {
					}
				}
				if (localDataCopy == null) {
					localDataCopy = new JSONObject();
				}
				callback.setLocalData(localDataCopy);
			}

			// Check transient callbacks
			if (callback == null && transientCallbacks != null) {
				callback = transientCallbacks.get(channel);
			}

			if (callback == null) {
				Log.d(TAG, "ignoring push on channel " + channel);
				return;
			}
		}

		// Handle the callback
		Log.d(TAG, "received push on channel " + channel);
		callback.setService(this);
		callback.setPushData(pushData);
		callback.setChannel(channel);
		callback.run();
	}	
	
	static void sleep(int millis) throws InterruptedException {
		Semaphore semaphore = sleepSemaphore;
		if (semaphore == null) {
			Thread.sleep(millis);
		} else {
			// Sleep in units of 100 millis
			while (millis > 100) {
				semaphore.acquire(100);
				millis -= 100;
			}
			semaphore.acquire(millis);
		}
	}
	
	static void useServer(String newServer) {
		pushServer = newServer;
	}
	
	static void usePort(int newPort) {
		defaultPushPort = newPort;
	}
	
	static void resetDelay() {
		delaySeconds = 0;
		consecutiveFailures = 0;
	}
	
	// Gets the application id and caches it on disk if it isn't there yet.
	private static synchronized String getApplicationId(Context context) {
		JSONObject oauth = ParseObject.getDiskObject(context, "oauth");
		if (oauth == null) {
			oauth = new JSONObject();
		}
		String cached = oauth.optString("key");
		if (cached != "") {
			return cached;
		}
		String applicationId = ParseObject.getApplicationId();
		try {
			oauth.put("key", applicationId);
		} catch (JSONException e) {
			Log.e(TAG, e.getMessage());
		}
		ParseObject.saveDiskObject(context, "oauth", oauth);
		return applicationId;
	}
	
	// Increase the delay but delay at most five minutes
	private void increaseDelay() {
		++consecutiveFailures;
		delaySeconds *= 2;
		delaySeconds = Math.max(1, delaySeconds);
		delaySeconds = Math.min(delaySeconds, 60 * 5);
	}
	
	/**
	 * Client code should not call onCreate directly.
	 */
    @Override
    public void onCreate() {
    	super.onCreate();
    	Log.d(TAG, "creating push service");
    	pushPort = defaultPushPort;
    	keepAliveTimer = new Timer("com.parse.PushService.keepAliveTimer", true);
    	long msKeepAlive = 1000 * 60 * 20; // 20 minutes
    	keepAliveTimer.schedule(new TimerTask() {
			@Override
			public void run() {
				PushService.this.sendKeepAlive();
			}
    	}, msKeepAlive, msKeepAlive);
    	resetDelay();
    	
		readInBackground(null);
    }
    
    // If there is no open socket, this just does nothing.
    // The context is required to access cached subscription information.
    // Returns whether the send seemed to work.
    // This function can block the thread trying to write to the socket.
    private static boolean sendSubscriptionInformation(Context context) {
    	if (socket == null) {
    		return false;
    	}

    	// Make a set of channels by merging transient and persistent channels
    	Set<String> channelSet = new HashSet<String>();
    	synchronized (PushCallback.class) {
    		if (transientCallbacks != null) {
    			for (String channel : transientCallbacks.keySet()) {
    				channelSet.add(channel);
    			}
    		}
    		loadPersistentCallbacks(context);
    		@SuppressWarnings("unchecked")
    		Iterator<String> iter = persistentCallbacks.keys();
    		while (iter.hasNext()) {
    			String channel = iter.next();
    			channelSet.add(channel);
    		}
    	}

		// Convert channels to json
		JSONArray channels = new JSONArray();
		for (String channel : channelSet) {
			channels.put(channel);
		}
		
		JSONObject subscription = new JSONObject();
		try {
			subscription.put("installation_id", ParseObject.getInstallationId(context));
			subscription.put("oauth_key", getApplicationId(context));
			subscription.put("subscribe", channels);
			if (lastTime == null) {
				subscription.put("last", JSONObject.NULL);
			} else {
				subscription.put("last", lastTime);
			}
		} catch (JSONException e) {
			Log.d(TAG, "unexpected JSONException");
			return false;
		}
		try {
			BufferedWriter writer = new BufferedWriter(
					new OutputStreamWriter(socket.getOutputStream(), "UTF-8"),
					8 * 1024);
			String message = subscription.toString();
			Log.d(TAG, "subscribing with " + message + " @ " + socket.getPort());
			writer.write(message + "\n");
			writer.flush();
		} catch (UnsupportedEncodingException e) {
			Log.d(TAG, "unsupported encoding");
			return false;
		} catch (IOException e) {
			Log.d(TAG, "could not construct writer: " + e.getMessage());
			return false;
		}
		return true;
    }
    
    // Sends a packet for the purposes of keepalives.
    // This can block the thread trying to send.
    // TODO: ensure threadsafety
    private void sendKeepAlive() {
    	if (socket == null) {
    		return;
    	}
		try {
			BufferedWriter writer = new BufferedWriter(
					new OutputStreamWriter(socket.getOutputStream(), "UTF-8"),
					8 * 1024);
			writer.write("{}\n");
			writer.flush();
		} catch (UnsupportedEncodingException e) {
			Log.d(TAG, "unsupported encoding in keepalive");
			return;
		} catch (IOException e) {
			Log.d(TAG, "could not construct writer in keepalive: " + e.getMessage());
			return;
		}
    }
    
    // Uses a background thread to read from the socket and repeat.
    // This gets passed a BufferedReader if there is one.
    // If the reader is null, this constructs a new socket and a new reader.
    // 
    // When the PushService is being shut down for low memory conditions,
    // there are no guarantees for what happens to this thread.
    // onDestroy might be called at any time, and it should just stop
    // what this thread is doing, without actually killing it, because
    // that would mess with the ParseAsyncTask threadpool.
    // 
    // To handle this, we check destroyed after every place
    // that is likely to yield to another thread. This is recommended by
    // the Android docs for cleaning up AsyncTasks, but it clearly is not ideal.
    private void readInBackground(final BufferedReader initialReader) {
    	final Service finalService = this;
    	if (destroyed) {
    		return;
    	}
    	task = new ParseAsyncTask<Void, Void, String>() {
    		private BufferedReader reader;
    		
    		@Override
			protected String doInBackground(Void... v) {
    			if (initialReader != null) {
    				reader = initialReader;
    			} else {
    				if (delaySeconds > 0) {
    					try {
    						Log.d(TAG, "sleeping for " + delaySeconds + " seconds");
    						sleep(delaySeconds * 1000);
    					} catch (InterruptedException e) {
    						Log.d(TAG, "sleep interrupted");
    						return null;
    					}
    				}
    				if (destroyed) {
    					return null;
    				}
    				increaseDelay();

    				// Close the old socket
    				if (socket != null) {
    					try {
    						socket.close();
    					} catch (IOException e) {
    						// If we can't close the old socket, just ignore it.
    						Log.d(TAG, "error closing socket: " + e.getMessage());
    					}
    				}

    				// Construct the new socket before assigning it.
    				Socket newSocket;
    				try {
    					Log.d(TAG, "connecting to push server at " +
    							pushServer + ":" + pushPort);
    					InetSocketAddress address = new InetSocketAddress(pushServer, pushPort);
    					newSocket = new Socket();
    					newSocket.setKeepAlive(true);
    					if (sleepSemaphore == null) {
    						newSocket.connect(address, 5000);
    					} else {
    						// Speed up unit tests
    						newSocket.connect(address, 50);
    					}
    				} catch (UnknownHostException e) {
    					Log.d(TAG, "unknown host");
    					return null;
    				} catch (IOException e) {
    					Log.d(TAG, "could not connect to push server");
    					return null;
    				}

    				// Synchronize to prevent race conditions between
    				// socket creation and socket destruction.
    				synchronized (PushService.this) {
    					if (destroyed) {
    						return null;
    					}
    					socket = newSocket;
    				}
    				
					// Create a new reader
					try {
						reader = new BufferedReader(
								new InputStreamReader(socket.getInputStream(), "UTF-8"),
								8 * 1024);
					} catch (UnsupportedEncodingException e) {
						Log.d(TAG, "unsupported encoding");
						return null;
					} catch (IOException e) {
						Log.d(TAG, "could not construct reader");
						return null;
					}

					if (!sendSubscriptionInformation(finalService)) {
						return null;
					}
					
					if (destroyed) {
						return null;
					}
    			}
    			
    			resetDelay();
    			
    			try {
    				// This will block for perhaps a very long time
    				Log.d(TAG, "waiting for a push");
    				String line = reader.readLine();
					return line;
				} catch (IOException e) {
					Log.d(TAG, "could not read from socket: " + e.getMessage());
					return null;
				}
    		}

			@Override
			protected void onPostExecute(String line) {
				if (destroyed) {
					return;
				}
				if (line == null) {					
					readInBackground(null);
					return;
				}
				
				// We need to extract the channel and datafrom the json result
				JSONTokener tokener = new JSONTokener(line);
				JSONObject message;
				try {
					message = new JSONObject(tokener);
				} catch (JSONException e) {
					// The json was bad but we can still use the socket
					Log.d(TAG, "bad json: " + line);
					readInBackground(reader);
					return;
				}

				// Update lastTime
				if (message.has("time")) {
					String time = message.optString("time");
					if (lastTime == null || time.compareTo(lastTime) > 0) {
						lastTime = time;
					}
				}				
				
				// Handle the push data
				if (message.has("channel")) {
					String channel = message.optString("channel");
					JSONObject pushData = message.optJSONObject("data");
					handlePushData(channel, pushData);
				}
				
				readInBackground(reader);
			}
    	};
    	task.execute();
    }

    /**
     * onBind should not be called directly.
     */
	@Override
	public IBinder onBind(Intent intent) {
		throw new IllegalArgumentException(
				"You cannot bind directly to the PushService. " +
				"Use PushService.subscribe instead.");
	}
	
	/**
	 * Client code should not call onDestroy directly.
	 */
	@Override
	public void onDestroy() {
		// We never explicitly stop the service, since it is meant to always keep running
		// in the background, so this should only end up being called when the system is
		// running low on memory or we are unit testing.	
		super.onDestroy();
		Log.d(TAG, "destroying push service");
		task.cancel(true);
		keepAliveTimer.cancel();
		
		// Synchronize to avoid race conditions between socket cleanup
		// and socket creation
		Socket oldSocket;
		synchronized (this) {
			destroyed = true;
			oldSocket = socket;
			socket = null;
		}
		if (oldSocket != null) {
			try {
				oldSocket.close();
			} catch (IOException e) {
				// Not much we can do here
			}
		}
	}
}
