package com.parse;

/**
 * A RefreshCallback is used to run code after refresh is used to update a {@link ParseObject} in
 * a background thread.
 * <p>
 * The easiest way to use a RefreshCallback is through an anonymous inner class. Override the
 * <code>done</code> function to specify what the callback should do after the refresh is complete. The
 * <code>done</code> function will be run in the UI thread, while the refresh happens in a background thread.
 * This ensures that the UI does not freeze while the refresh happens. 
 * <p>
 * For example, this sample code refreshes an object of class <code>"MyClass"</code> and id <code>myId</code>.
 * It calls a different function depending on whether the refresh succeeded or not.
 * <p>
 * <pre>
 * object.refreshInBackground(new RefreshCallback() {
 *     public void done(ParseObject object, ParseException e) {
 *         if (e == null) {
 *             objectWasRefreshedSuccessfully(object);
 *         } else {
 *             objectRefreshFailed();
 *         }
 *     }
 * });
 * </pre>
 */
public abstract class RefreshCallback extends ParseCallback<ParseObject> {

	/**
	 * Override this function with the code you want to run after
	 * the save is complete.
	 * @param e The exception raised by the save, or null if it succeeded.
	 */
	public abstract void done(ParseObject object, ParseException e);
	
	@Override
	final void internalDone(ParseObject returnValue, ParseException e) {
		done(returnValue, e);
	}
}
