package com.parse;

/**
 * A RequestPasswordResetCallback is used to run code requesting
 * a password reset for a user.
 * 
 * <p>
 * The easiest way to use a RequestPasswordResetCallback is through an anonymous inner class. Override the
 * <code>done</code> function to specify what the callback should do after the request is complete. The
 * <code>done</code> function will be run in the UI thread, while the request happens in a background thread.
 * This ensures that the UI does not freeze while the request happens. 
 * <p>
 * For example, this sample code requests a password reset for a user and calls a different function
 * depending on whether the request succeeded or not.
 * <p>
 * <pre>
 * ParseUser.requestPasswordResetInBackground("forgetful@example.com", new RequestPasswordResetCallback() {
 *     public void done(ParseException e) {
 *         if (e == null) {
 *             requestedSuccessfully();
 *         } else {
 *             requestDidNotSucceed();
 *         }
 *     }
 * });
 * </pre>
 */
public abstract class RequestPasswordResetCallback extends ParseCallback<Void> {

	/**
	 * Override this function with the code you want to run after
	 * the request is complete.
	 * @param e The exception raised by the save, or null if no account is
	 * associated with the email address.
	 */
	public abstract void done(ParseException e);
	
	@Override
	final void internalDone(Void returnValue, ParseException e) {
		done(e);
	}
}
