package com.parse;

/**
 * A SaveCallback is used to run code after saving a {@link ParseObject} in a background thread.
 * <p>
 * The easiest way to use a SaveCallback is through an anonymous inner class. Override the
 * <code>done</code> function to specify what the callback should do after the save is complete. The
 * <code>done</code> function will be run in the UI thread, while the save happens in a background thread.
 * This ensures that the UI does not freeze while the save happens. 
 * <p>
 * For example, this sample code saves the object <code>myObject</code> and calls a different function
 * depending on whether the save succeeded or not.
 * <p>
 * <pre>
 * myObject.saveInBackground(new SaveCallback() {
 *     public void done(ParseException e) {
 *         if (e == null) {
 *             myObjectSavedSuccessfully();
 *         } else {
 *             myObjectSaveDidNotSucceed();
 *         }
 *     }
 * });
 * </pre>
 */
public abstract class SaveCallback extends ParseCallback<Void> {

	/**
	 * Override this function with the code you want to run after
	 * the save is complete.
	 * @param e The exception raised by the save, or null if it succeeded.
	 */
	public abstract void done(ParseException e);
	
	@Override
	final void internalDone(Void returnValue, ParseException e) {
		done(e);
	}

}
