package com.parse;

/**
 * A SendCallback is used to run code after sending a {@link ParsePush} in a background thread.
 * <p>
 * The easiest way to use a SendCallback is through an anonymous inner class. Override the
 * <code>done</code> function to specify what the callback should do after the send is complete. The
 * <code>done</code> function will be run in the UI thread, while the send happens in a background thread.
 * This ensures that the UI does not freeze while the send happens. 
 * <p>
 * For example, this sample code sends the message <code>"Hello world"</code> on the
 * <code>"hello"</code> channel and logs whether the send succeeded.
 * <p>
 * <pre>
 * ParsePush push = new ParsePush();
 * push.setChannel("hello");
 * push.setMessage("Hello world!");
 * push.sendInBackground(new SendCallback() {
 *     public void done(ParseException e) {
 *         if (e == null) {
 *             Log.d("push", "success!");
 *         } else {
 *             Log.d("push", "failure");
 *         }
 *     }
 * });
 * </pre>
 */
public abstract class SendCallback extends ParseCallback<Void> {

	/**
	 * Override this function with the code you want to run after
	 * the send is complete.
	 * @param e The exception raised by the send, or null if it succeeded.
	 */
	public abstract void done(ParseException e);
		
	@Override
	final void internalDone(Void returnValue, ParseException e) {
		done(e);
	}
}
