package com.parse;

/**
 * A SignUpCallback is used to run code after signing up a {@link ParseUser} in a background thread.
 * <p>
 * The easiest way to use a SignUpCallback is through an anonymous inner class. Override the
 * <code>done</code> function to specify what the callback should do after the save is complete. The
 * <code>done</code> function will be run in the UI thread, while the signup happens in a background thread.
 * This ensures that the UI does not freeze while the signup happens. 
 * <p>
 * For example, this sample code signs up the object <code>myUser</code> and calls a different function
 * depending on whether the signup succeeded or not.
 * <p>
 * <pre>
 * myUser.signUpInBackground(new SignUpCallback() {
 *     public void done(ParseException e) {
 *         if (e == null) {
 *             myUserSignedUpSuccessfully();
 *         } else {
 *             myUserSignUpDidNotSucceed();
 *         }
 *     }
 * });
 * </pre>
 */
public abstract class SignUpCallback extends ParseCallback<Void> {

	/**
	 * Override this function with the code you want to run after
	 * the signUp is complete.
	 * @param e The exception raised by the signUp, or null if it succeeded.
	 */
	public abstract void done(ParseException e);
	
	@Override
	final void internalDone(Void returnValue, ParseException e) {
		done(e);
	}

}
